# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2015-2019  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Base class for everything Codimension puts on the graphics scene"""

# pylint: disable=C0305

from sys import maxsize
from ui.qt import QMimeData, Qt, QApplication, QPen
from utils.globals import GlobalData
from utils.config import DEFAULT_ENCODING
from .cml import CMLVersion
from .routines import distance


class CellElement:

    """Base class for all the elements which could be found on the canvas"""

    UNKNOWN = -1

    VCANVAS = 0

    VACANT = 1
    H_SPACER = 2
    V_SPACER = 3
    H_GROUP_SPACER = 4
    SPACER = 5

    NO_SCOPE = 99
    FILE_SCOPE = 100
    FUNC_SCOPE = 101
    CLASS_SCOPE = 102
    FOR_SCOPE = 103
    WHILE_SCOPE = 104
    TRY_SCOPE = 105
    WITH_SCOPE = 106
    DECOR_SCOPE = 107
    ELSE_SCOPE = 108
    EXCEPT_SCOPE = 109
    FINALLY_SCOPE = 110
    SCOPE_H_SIDE_EDGE = 111
    SCOPE_V_SIDE_EDGE = 112
    SCOPE_CORNER_EDGE = 113

    # The three items below will be mapped to ELSE_SCOPE
    # They are needed to distinguish the constructors which use individual
    # color settings for each kind of else scope. I.e. the graphics scene
    # will have no items of these types ever
    FOR_ELSE_SCOPE = 120
    WHILE_ELSE_SCOPE = 121
    TRY_ELSE_SCOPE = 122

    CODE_BLOCK = 200
    BREAK = 201
    CONTINUE = 202
    RETURN = 203
    RAISE = 204
    ASSERT = 205
    SYSEXIT = 206
    IMPORT = 207
    IF = 208
    LEADING_COMMENT = 209
    INDEPENDENT_COMMENT = 210
    SIDE_COMMENT = 211
    ABOVE_COMMENT = 212
    EXCEPT_MINIMIZED = 213
    INDEPENDENT_DOC = 214
    LEADING_DOC = 215
    ABOVE_DOC = 216
    INDEPENDENT_MINIMIZED_COMMENT = 217
    LEADING_MINIMIZED_COMMENT = 218
    SIDE_MINIMIZED_COMMENT = 219
    ABOVE_MINIMIZED_COMMENT = 220

    CONNECTOR = 300
    SVG = 301
    BADGE = 302
    DEPENDENT_CONNECTOR = 303
    TEXT = 304
    RUBBER_BAND = 305
    LINE = 306
    RECTANGLE = 307
    GROUP_CORNER_CONROL = 308

    EMPTY_GROUP = 500
    OPENED_GROUP_BEGIN = 501
    OPENED_GROUP_END = 502
    COLLAPSED_GROUP = 503

    def __init__(self, ref, canvas, x, y):
        self.kind = self.UNKNOWN
        self.subKind = self.UNKNOWN
        self.ref = ref              # reference to the control flow object
        self.addr = [x, y]          # indexes in the current canvas
        self.canvas = canvas        # reference to the canvas

        self.tailComment = False

        # Filled when rendering is called
        self.width = None
        self.height = None
        self.minWidth = None
        self.minHeight = None

        # Filled when draw is called
        self.baseX = None
        self.baseY = None

        # Shift is used when open groups are involved
        self.hShift = 0

        # Unique sequential ID
        self.itemID = canvas.settings.itemID
        canvas.settings.itemID += 1

    def __str__(self):
        return kindToString(self.kind) + \
            '[' + str(self.minWidth) + 'x' + str(self.minHeight) + '] -> [' + \
            str(self.width) + 'x' + str(self.height) + ']'

    def render(self):
        """Renders the graphics considering settings"""
        raise Exception("render() is not implemented for " +
                        kindToString(self.kind))

    def draw(self, scene, baseX, baseY):
        """Draws the element on the real canvas. Should respect settings."""
        del scene   # unused argument
        del baseX   # unused argument
        del baseY   # unused argument
        raise Exception("draw() is not implemented for " +
                        kindToString(self.kind))

    def getBoundingRect(self, text):
        """Provides the bounding rectangle for a monospaced font"""
        return self.canvas.settings.monoFontMetrics.boundingRect(
            0, 0, maxsize, maxsize, 0, text)

    def getTooltip(self):
        """Provides the tooltip"""
        parts = []
        canvas = self.canvas
        while canvas is not None:
            parts.insert(0, canvas.getScopeName())
            canvas = canvas.canvas
        if self.canvas.settings.debug:
            return "::".join(parts) + "<br>Size: " + \
                str(self.width) + "x" + str(self.height) + \
                " (" + str(self.minWidth) + "x" + str(self.minHeight) + ")" + \
                " Row: " + str(self.addr[1]) + " Column: " + str(self.addr[0])
        return "::".join(parts)

    def getCanvasTooltip(self):
        """Provides the canvas tooltip"""
        parts = []
        canvas = self.canvas
        while canvas is not None:
            if not canvas.isNoScope:
                parts.insert(0, canvas.getScopeName())
            canvas = canvas.canvas
        path = " :: ".join(parts)
        if not path:
            path = " ::"
        if self.canvas.settings.debug:
            return path + "<br>Size: " + str(self.canvas.width) + "x" + \
                   str(self.canvas.height) + \
                   " (" + str(self.canvas.minWidth) + "x" + \
                   str(self.canvas.minHeight) + ")"
        return path

    def getEditor(self):
        """Provides a reference to the editor"""
        canvas = self.canvas
        while canvas.canvas is not None:
            canvas = canvas.canvas
        return canvas.editor

    def getPainterPen(self, selected, borderColor):
        """Provides the painter pen for the item"""
        if selected:
            pen = QPen(self.canvas.settings.selectColor)
            pen.setWidth(self.canvas.settings.selectPenWidth)
        else:
            pen = QPen(borderColor)
            pen.setWidth(self.canvas.settings.boxLineWidth)
        pen.setJoinStyle(Qt.RoundJoin)
        return pen

    def mouseDoubleClickEvent(self, event, line=None, pos=None):
        """Jump to the appropriate line in the text editor.

        default implementation
        """
        editor = self.getEditor()
        if editor is not None:
            if event:
                if event.buttons() != Qt.LeftButton:
                    return

            GlobalData().mainWindow.raise_()
            GlobalData().mainWindow.activateWindow()

            if line is None:
                line = self.getLineRange()[0]
            if pos is None:
                lineContent = editor.lines[line - 1]
                pos = len(lineContent) - len(lineContent.lstrip()) + 1

            editor.gotoLine(line, pos)
            editor.setFocus()

    def scopedItem(self):
        """True if it is a scoped item"""
        return self.kind in (self.FILE_SCOPE, self.FUNC_SCOPE,
                             self.CLASS_SCOPE, self.FOR_SCOPE,
                             self.WHILE_SCOPE, self.TRY_SCOPE,
                             self.WITH_SCOPE, self.DECOR_SCOPE,
                             self.ELSE_SCOPE, self.EXCEPT_SCOPE,
                             self.FINALLY_SCOPE,
                             self.SCOPE_H_SIDE_EDGE, self.SCOPE_V_SIDE_EDGE,
                             self.SCOPE_CORNER_EDGE)

    def isProxyItem(self):
        """True if it is a proxy item"""
        return self.kind in (self.BADGE, self.SVG, self.CONNECTOR,
                             self.DEPENDENT_CONNECTOR, self.TEXT,
                             self.RUBBER_BAND, self.LINE, self.RECTANGLE,
                             self.GROUP_CORNER_CONROL)

    def isSpacerItem(self):
        """True is it is a some kind of a spacer"""
        return self.kind in (self.VACANT, self.H_SPACER, self.V_SPACER,
                             self.H_GROUP_SPACER, self.SPACER)

    @staticmethod
    def getProxiedItem():
        """Provides the real item for a proxy one"""
        return None

    def isComment(self):
        """True if it is a comment"""
        return self.kind in (self.INDEPENDENT_COMMENT,
                             self.LEADING_COMMENT,
                             self.SIDE_COMMENT,
                             self.ABOVE_COMMENT,
                             self.INDEPENDENT_MINIMIZED_COMMENT,
                             self.LEADING_MINIMIZED_COMMENT,
                             self.SIDE_MINIMIZED_COMMENT,
                             self.ABOVE_MINIMIZED_COMMENT)

    def isCMLDoc(self):
        """True if it is a CML doc item"""
        return self.kind in (self.INDEPENDENT_DOC, self.LEADING_DOC,
                             self.ABOVE_DOC)

    def isGroupItem(self):
        """True if it is some kind of a group item"""
        return self.kind in (self.OPENED_GROUP_BEGIN, self.OPENED_GROUP_END,
                             self.COLLAPSED_GROUP, self.EMPTY_GROUP)

    @staticmethod
    def isDocstring():
        """True if it is a docstring"""
        return False

    def isMinimizedItem(self):
        """True if it is a minimized item"""
        return self.kind in (self.INDEPENDENT_MINIMIZED_COMMENT,
                             self.LEADING_MINIMIZED_COMMENT,
                             self.EXCEPT_MINIMIZED,
                             self.SIDE_MINIMIZED_COMMENT,
                             self.ABOVE_MINIMIZED_COMMENT)

    def getDistance(self, absPos):
        """Default implementation.

        Provides a distance between the absPos and the item
        """
        absPosRange = self.getAbsPosRange()
        return distance(absPos, absPosRange[0], absPosRange[1])

    def getLineDistance(self, line):
        """Default implementation.

        Provides a distance between the line and the item
        """
        lineRange = self.getLineRange()
        return distance(line, lineRange[0], lineRange[1])

    def getFirstLine(self):
        """Provides the first line"""
        line = maxsize
        if hasattr(self.ref, "leadingCMLComments"):
            if self.ref.leadingCMLComments:
                line = CMLVersion.getFirstLine(self.ref.leadingCMLComments)
        if hasattr(self.ref, "leadingComment"):
            if self.ref.leadingComment:
                if self.ref.leadingComment.parts:
                    line = min(self.ref.leadingComment.parts[0].beginLine,
                               line)
        return min(self.ref.body.beginLine, line)

    def getLineRange(self):
        """Default implementation of the line range"""
        return self.ref.body.getLineRange()

    def getAbsPosRange(self):
        """Provides the absolute position range"""
        return [self.ref.body.begin, self.ref.body.end]

    @staticmethod
    def getLinesSuffix(lineRange):
        """Provides the proper suffix for the selected item tooltips"""
        if lineRange[0] == lineRange[1]:
            return 'line ' + str(lineRange[0])
        return 'lines ' + str(lineRange[0]) + "-" + str(lineRange[1])

    @staticmethod
    def _putMimeToClipboard(value):
        """Copies the value (string) to a clipboard as mime data"""
        mimeData = QMimeData()
        mimeData.setData('text/codimension', value.encode(DEFAULT_ENCODING))
        QApplication.clipboard().setMimeData(mimeData)

    def copyToClipboard(self):
        """Placeholder"""
        return


__kindToString = {
    CellElement.UNKNOWN: 'UNKNOWN',
    CellElement.VCANVAS: 'VCANVAS',
    CellElement.VACANT: 'VACANT',
    CellElement.H_SPACER: 'H_SPACER',
    CellElement.V_SPACER: 'V_SPACER',
    CellElement.H_GROUP_SPACER: 'H_GROUP_SPACER',
    CellElement.SPACER: 'SPACER',
    CellElement.NO_SCOPE: 'NO_SCOPE',
    CellElement.FILE_SCOPE: 'FILE_SCOPE',
    CellElement.FUNC_SCOPE: 'FUNC_SCOPE',
    CellElement.CLASS_SCOPE: 'CLASS_SCOPE',
    CellElement.DECOR_SCOPE: 'DECOR_SCOPE',
    CellElement.FOR_SCOPE: 'FOR_SCOPE',
    CellElement.WHILE_SCOPE: 'WHILE_SCOPE',
    CellElement.ELSE_SCOPE: 'ELSE_SCOPE',
    CellElement.WITH_SCOPE: 'WITH_SCOPE',
    CellElement.TRY_SCOPE: 'TRY_SCOPE',
    CellElement.EXCEPT_SCOPE: 'EXCEPT_SCOPE',
    CellElement.FINALLY_SCOPE: 'FINALLY_SCOPE',
    CellElement.SCOPE_H_SIDE_EDGE: 'SCOPE_H_SIDE_EDGE',
    CellElement.SCOPE_V_SIDE_EDGE: 'SCOPE_V_SIDE_EDGE',
    CellElement.SCOPE_CORNER_EDGE: 'SCOPE_CORNER_EDGE',
    CellElement.CODE_BLOCK: 'CODE_BLOCK',
    CellElement.BREAK: 'BREAK',
    CellElement.CONTINUE: 'CONTINUE',
    CellElement.RETURN: 'RETURN',
    CellElement.RAISE: 'RAISE',
    CellElement.ASSERT: 'ASSERT',
    CellElement.SYSEXIT: 'SYSEXIT',
    CellElement.IMPORT: 'IMPORT',
    CellElement.IF: 'IF',
    CellElement.LEADING_COMMENT: 'LEADING_COMMENT',
    CellElement.INDEPENDENT_COMMENT: 'INDEPENDENT_COMMENT',
    CellElement.SIDE_COMMENT: 'SIDE_COMMENT',
    CellElement.ABOVE_COMMENT: 'ABOVE_COMMENT',
    CellElement.INDEPENDENT_MINIMIZED_COMMENT: 'INDEPENDENT_MINIMIZED_COMMENT',
    CellElement.LEADING_MINIMIZED_COMMENT: 'LEADING_MINIMIZED_COMMENT',
    CellElement.SIDE_MINIMIZED_COMMENT: 'SIDE_MINIMIZED_COMMENT',
    CellElement.ABOVE_MINIMIZED_COMMENT: 'ABOVE_MINIMIZED_COMMENT',
    CellElement.EXCEPT_MINIMIZED: 'EXCEPT_MINIMIZED',
    CellElement.CONNECTOR: 'CONNECTOR',
    CellElement.DEPENDENT_CONNECTOR: 'DEPENDENT_CONNECTOR',
    CellElement.SVG: 'SVG',
    CellElement.BADGE: 'BADGE',
    CellElement.TEXT: 'TEXT',
    CellElement.RUBBER_BAND: 'RUBBER_BAND',
    CellElement.LINE: 'LINE',
    CellElement.RECTANGLE: 'RECTANGLE',
    CellElement.GROUP_CORNER_CONROL: 'GROUP_CORNER_CONROL',
    CellElement.INDEPENDENT_DOC: 'INDEPENDENT_DOC',
    CellElement.LEADING_DOC: 'LEADING_DOC',
    CellElement.ABOVE_DOC: 'ABOVE_DOC',
    CellElement.EMPTY_GROUP: 'EMPTY_GROUP',
    CellElement.OPENED_GROUP_BEGIN: 'OPENED_GROUP_BEGIN',
    CellElement.OPENED_GROUP_END: 'OPENED_GROUP_END',
    CellElement.COLLAPSED_GROUP: 'COLLAPSED_GROUP'}


def kindToString(kind):
    """Provides a string representation of a element kind"""
    return __kindToString[kind]




# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2016  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Find name feature implementation"""


from cdmpyparser import getBriefModuleInfoFromMemory
from utils.globals import GlobalData
from utils.pixmapcache import getIcon
from utils.settings import Settings
from utils.fileutils import isPythonFile
from utils.diskvaluesrelay import getFindNameHistory, setFindNameHistory
from .qt import (Qt, QAbstractItemModel, QRegExp, QModelIndex,
                 QTreeView, QAbstractItemView, QDialog, QVBoxLayout, QCursor,
                 QComboBox, QSizePolicy, QSortFilterProxyModel, QApplication)
from .combobox import EnterSensitiveComboBox
from .itemdelegates import NoOutlineHeightDelegate


class NameItem():

    """Names list item"""

    def __init__(self, parent, icon, name, fileName, line, tooltip):
        self.parentItem = parent
        self.icon = icon

        self.childItems = []
        self.childItemsSize = 0

        # Item data is below
        self.name = name
        self.tooltip = ""
        if parent is not None:
            if tooltip == "":
                self.tooltip = fileName + ":" + str(line)
            else:
                self.tooltip = tooltip + "\n\n" + fileName + ":" + str(line)

    @staticmethod
    def columnCount():
        """Provides the number of columns"""
        return 1

    def data(self, column):
        """Provides a value for the column"""
        if column == 0:
            return self.name
        return ''

    def appendChild(self, child):
        """Appends a child item"""
        self.childItems.append(child)
        self.childItemsSize += 1

    def childCount(self):
        """Provides the number of children"""
        return self.childItemsSize

    def removeChildren(self):
        """Removes all the children"""
        self.childItems = []
        self.childItemsSize = 0

    def child(self, row):
        """Provides a reference to a child"""
        return self.childItems[row]

    def parent(self):
        """Provides a reference to the parent item"""
        return self.parentItem

    # Arguments: other, column, order
    def lessThan(self, other, column, _):
        """Check, if the item is less than another"""
        if column == 0:
            return self.name < other.name
        return False


class FindNameModel(QAbstractItemModel):

    """Find name data model implementation"""

    def __init__(self, parent=None):
        QAbstractItemModel.__init__(self, parent)

        self.rootItem = NameItem(None, None, "Name", "", "", "")
        self.count = 0
        self.showTooltips = Settings()['findNameTooltips']
        self.__populateModel()

    def __populateModel(self):
        """Populates the list names model"""
        self.clear()

        # If a project is loaded then take the project
        # If not - the opened files
        if GlobalData().project.isLoaded():
            self.__populateFromProject()
        else:
            self.__populateFromOpened()

    def __populateFromProject(self):
        """Populates find name dialog from the project files"""
        mainWindow = GlobalData().mainWindow
        for fname in GlobalData().project.filesList:
            if isPythonFile(fname):
                widget = mainWindow.getWidgetForFileName(fname)
                if widget is None:
                    info = GlobalData().briefModinfoCache.get(fname)
                else:
                    info = getBriefModuleInfoFromMemory(
                        widget.getEditor().text)
                self.__populateInfo(info, fname)

    def __populateInfo(self, info, fname):
        """Populates parsed info from a python file"""
        for item in info.globals:
            newItem = NameItem(self.rootItem, getIcon('globalvar.png'),
                               item.name, fname, item.line, "")
            self.rootItem.appendChild(newItem)
            self.count += 1

        for klass in info.classes:
            self.__polulateClass(klass, "", fname)
        for func in info.functions:
            self.__populateFunction(func, "", fname)

    def __polulateClass(self, klass, prefix, fname):
        """Recursively populates the given class"""
        if klass.isPrivate():
            icon = getIcon('class_private.png')
        elif klass.isProtected():
            icon = getIcon('class_protected.png')
        else:
            icon = getIcon('class.png')

        tooltip = ""
        if self.showTooltips and klass.docstring is not None:
            tooltip = klass.docstring.text

        classItem = NameItem(self.rootItem, icon, prefix + klass.name,
                             fname, klass.line, tooltip)
        self.rootItem.appendChild(classItem)
        self.count += 1

        for item in klass.classAttributes:
            newItem = NameItem(self.rootItem, getIcon('attributes.png'),
                               prefix + item.name, fname, item.line, "")
            self.rootItem.appendChild(newItem)
            self.count += 1

        for item in klass.instanceAttributes:
            newItem = NameItem(self.rootItem, getIcon('attributes.png'),
                               prefix + item.name, fname, item.line, "")
            self.rootItem.appendChild(newItem)
            self.count += 1

        for item in klass.functions:
            self.__populateFunction(item, prefix + klass.name + ".", fname)
        for item in klass.classes:
            self.__polulateClass(item, prefix + klass.name + ".", fname)

    def __populateFunction(self, func, prefix, fname):
        """Recursively populates the given function"""
        tooltip = ""
        if self.showTooltips and func.docstring is not None:
            tooltip = func.docstring.text

        funcItem = NameItem(self.rootItem, getIcon('method.png'),
                            prefix + func.name, fname, func.line, tooltip)
        self.rootItem.appendChild(funcItem)
        self.count += 1

        for item in func.functions:
            self.__populateFunction(item, prefix + func.name + ".", fname)
        for item in func.classes:
            self.__polulateClass(item, prefix + func.name + ".", fname)

    def __populateFromOpened(self):
        """Populates the name dialog from the opened files"""
        mainWindow = GlobalData().mainWindow
        editorsManager = mainWindow.editorsManagerWidget.editorsManager
        for record in editorsManager.getTextEditors():
            # uuid = record[0]
            fname = record[1]
            widget = record[2]
            if isPythonFile(fname):
                content = widget.getEditor().text()
                info = getBriefModuleInfoFromMemory(content)
                self.__populateInfo(info, fname)

    def columnCount(self, parent=QModelIndex()):
        """Provides the number of columns"""
        if parent.isValid():
            return parent.internalPointer().columnCount()
        return self.rootItem.columnCount()

    def rowCount(self, parent=QModelIndex()):
        """Provides the number of rows"""
        # Only the first column should have children
        if parent.column() > 0:
            return 0

        if not parent.isValid():
            return self.rootItem.childCount()

        parentItem = parent.internalPointer()
        return parentItem.childCount()

    def data(self, index, role):
        """Provides data of an item"""
        if not index.isValid():
            return None

        if role == Qt.DisplayRole:
            item = index.internalPointer()
            if index.column() < item.columnCount():
                return item.data(index.column())
            if index.column() == item.columnCount() and \
                 index.column() < self.columnCount(self.parent(index)):
                # This is for the case when an item under a multi-column
                # parent doesn't have a value for all the columns
                return ""
        elif role == Qt.DecorationRole:
            if index.column() == 0:
                return index.internalPointer().icon
        elif role == Qt.ToolTipRole:
            item = index.internalPointer()
            if item.tooltip != "":
                return item.tooltip
        return None

    @staticmethod
    def flags(index):
        """Provides the item flags"""
        if not index.isValid():
            return Qt.ItemIsEnabled
        return Qt.ItemIsEnabled | Qt.ItemIsSelectable

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        """Provides the header data"""
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            if section >= self.rootItem.columnCount():
                return ""
            return self.rootItem.data(section)
        return None

    def index(self, row, column, parent=QModelIndex()):
        """Creates an index"""
        # The model/view framework considers negative values out-of-bounds,
        # however in python they work when indexing into lists. So make sure
        # we return an invalid index for out-of-bounds row/col
        if row < 0 or column < 0 or \
           row >= self.rowCount(parent) or \
           column >= self.columnCount(parent):
            return QModelIndex()

        if not parent.isValid():
            parentItem = self.rootItem
        else:
            parentItem = parent.internalPointer()

        try:
            childItem = parentItem.child(row)
        except IndexError:
            childItem = None
            return QModelIndex()

        if childItem:
            return self.createIndex(row, column, childItem)
        return QModelIndex()

    def parent(self, index):
        """Provides the index of the parent object"""
        if not index.isValid():
            return QModelIndex()

        childItem = index.internalPointer()
        parentItem = childItem.parent()

        if parentItem == self.rootItem:
            return QModelIndex()

        return self.createIndex(parentItem.row(), 0, parentItem)

    def hasChildren(self, parent=QModelIndex()):
        """Checks for the presence of child items"""
        # Only the first column should have children
        if parent.column() > 0:
            return False

        if not parent.isValid():
            return self.rootItem.childCount() > 0
        return parent.internalPointer().childCount() > 0

    def clear(self):
        """Clears the model"""
        self.beginResetModel()
        self.rootItem.removeChildren()
        self.endResetModel()

    @staticmethod
    def item(index):
        """Provides a reference to an item"""
        if not index.isValid():
            return None
        return index.internalPointer()


class FindNameSortFilterProxyModel(QSortFilterProxyModel):

    """Find name dialog sort filter proxy model"""

    def __init__(self, parent=None):
        QSortFilterProxyModel.__init__(self, parent)
        self.__sortColumn = None    # Avoid pylint complains
        self.__sortOrder = None     # Avoid pylint complains

        self.__filters = []
        self.__filtersCount = 0
        self.__sourceModelRoot = None

    def sort(self, column, order):
        """Sorts the items"""
        self.__sortColumn = column
        self.__sortOrder = order
        QSortFilterProxyModel.sort(self, column, order)

    def lessThan(self, left, right):
        """Sorts the displayed items"""
        lhs = left.model() and left.model().item(left) or None
        rhs = right.model() and right.model().item(right) or None

        if lhs and rhs:
            return lhs.lessThan(rhs, self.__sortColumn, self.__sortOrder)
        return False

    def item(self, index):
        """Provides a reference to the item"""
        if not index.isValid():
            return None

        sourceIndex = self.mapToSource(index)
        return self.sourceModel().item(sourceIndex)

    def hasChildren(self, parent=QModelIndex()):
        """Checks the presence of the child items"""
        sourceIndex = self.mapToSource(parent)
        return self.sourceModel().hasChildren(sourceIndex)

    def setFilter(self, text):
        """Sets the new filters"""
        self.__filters = []
        self.__filtersCount = 0
        self.__sourceModelRoot = None
        for part in str(text).strip().split():
            regexp = QRegExp(part, Qt.CaseInsensitive, QRegExp.RegExp2)
            self.__filters.append(regexp)
            self.__filtersCount += 1
        self.__sourceModelRoot = self.sourceModel().rootItem

    # Arguments: sourceRow, sourceParent
    def filterAcceptsRow(self, sourceRow, _):
        """Filters rows"""
        if self.__filtersCount == 0 or self.__sourceModelRoot is None:
            return True     # No filters

        nameToMatch = self.__sourceModelRoot.child(sourceRow).name
        for regexp in self.__filters:
            if regexp.indexIn(nameToMatch) == -1:
                return False
        return True


class NamesBrowser(QTreeView):

    """List of names widget implementation"""

    def __init__(self, parent=None):
        QTreeView.__init__(self, parent)

        self.__parentDialog = parent
        self.__model = FindNameModel()
        self.__sortModel = FindNameSortFilterProxyModel()
        self.__sortModel.setDynamicSortFilter(True)
        self.__sortModel.setSourceModel(self.__model)
        self.setModel(self.__sortModel)
        self.selectedIndex = None

        self.activated.connect(self.openCurrentItem)

        self.setRootIsDecorated(False)
        self.setAlternatingRowColors(True)
        self.setUniformRowHeights(True)
        self.setItemDelegate(NoOutlineHeightDelegate(4))

        header = self.header()
        header.setSortIndicator(0, Qt.AscendingOrder)
        header.setSortIndicatorShown(True)
        header.setSectionsClickable(True)

        self.setSortingEnabled(True)

        self.setSelectionMode(QAbstractItemView.SingleSelection)
        self.setSelectionBehavior(QAbstractItemView.SelectRows)

        self.layoutDisplay()

    def selectionChanged(self, selected, deselected):
        """Slot is called when the selection has been changed"""
        if selected.indexes():
            self.selectedIndex = selected.indexes()[0]
        else:
            self.selectedIndex = None
        QTreeView.selectionChanged(self, selected, deselected)

    def layoutDisplay(self):
        """Performs the layout operation"""
        self.header().setStretchLastSection(True)
        self._resort()

    def _resort(self):
        """Re-sorts the tree"""
        self.model().sort(self.header().sortIndicatorSection(),
                          self.header().sortIndicatorOrder())

    def openCurrentItem(self):
        """Triggers when an item is clicked or double clicked"""
        if self.selectedIndex is not None:
            item = self.model().item(self.selectedIndex)
            self.openItem(item)

    def openItem(self, item):
        """Handles the case when an item is activated"""
        infoLine = item.tooltip.split('\n')[-1]
        parts = infoLine.split(':')
        line = int(parts[-1])
        fname = ':'.join(parts[:-1])

        GlobalData().mainWindow.openFile(fname, line)
        self.__parentDialog.onClose()

    def getTotal(self):
        """Provides the total number of items"""
        return self.model().sourceModel().count

    def getVisible(self):
        """Provides the number of visible items"""
        return self.model().rowCount()

    def setFilter(self, text):
        """Called when the filter has been changed"""
        # Notify the filtering model of the new filters
        self.model().setFilter(text)

        # This is to trigger filtering - ugly but I don't know how else
        self.model().setFilterRegExp("")


class FindNameDialog(QDialog):

    """Find name dialog implementation"""

    def __init__(self, what="", parent=None):
        QDialog.__init__(self, parent)

        self.__namesBrowser = None
        self.findCombo = None
        self.__projectLoaded = GlobalData().project.isLoaded()

        QApplication.setOverrideCursor(QCursor(Qt.WaitCursor))
        self.__createLayout()
        self.findCombo.setFocus()
        QApplication.restoreOverrideCursor()

        # Set the window title and restore the previous searches
        self.__findNameHistory = getFindNameHistory()
        self.findCombo.addItems(self.__findNameHistory)
        self.findCombo.setEditText(what)

        self.findCombo.editTextChanged.connect(self.__filterChanged)

        self.__highlightFirst()
        self.__updateTitle()

    def __highlightFirst(self):
        """Sets the selection to the first item in the files list"""
        if self.__namesBrowser.getVisible() == 0:
            return
        self.__namesBrowser.clearSelection()

        first = self.__namesBrowser.model().index(0, 0, QModelIndex())
        self.__namesBrowser.setCurrentIndex(first)
        self.__namesBrowser.scrollTo(first)

    def __updateTitle(self):
        """Updates the window title"""
        title = "Find name in the "
        if self.__projectLoaded:
            title += "project: "
        else:
            title += "opened files: "
        title += str(self.__namesBrowser.getVisible()) + " of " + \
                 str(self.__namesBrowser.getTotal())
        self.setWindowTitle(title)

    def __createLayout(self):
        """Creates the dialog layout"""
        self.resize(600, 300)
        self.setSizeGripEnabled(True)

        verticalLayout = QVBoxLayout(self)
        self.__namesBrowser = NamesBrowser(self)
        verticalLayout.addWidget(self.__namesBrowser)

        self.findCombo = EnterSensitiveComboBox(self)
        self.__tuneCombo(self.findCombo)
        self.findCombo.lineEdit().setToolTip("Regular expression "
                                             "to search for")
        verticalLayout.addWidget(self.findCombo)
        self.findCombo.enterClicked.connect(self.__enterInFilter)

    @staticmethod
    def __tuneCombo(comboBox):
        """Sets the common settings for a combo box"""
        sizePolicy = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            comboBox.sizePolicy().hasHeightForWidth())
        comboBox.setSizePolicy(sizePolicy)
        comboBox.setEditable(True)
        comboBox.setInsertPolicy(QComboBox.InsertAtTop)
        comboBox.setCompleter(None)
        comboBox.setDuplicatesEnabled(False)

    def __filterChanged(self, text):
        """Triggers when the filter text changed"""
        self.__namesBrowser.setFilter(text)
        self.__highlightFirst()
        self.__updateTitle()

    def onClose(self):
        """Called when an item has been selected"""
        # Save the current filter if needed
        filterText = self.findCombo.currentText().strip()
        if filterText != "":
            if filterText in self.__findNameHistory:
                self.__findNameHistory.remove(filterText)
            self.__findNameHistory.insert(0, filterText)
            if len(self.__findNameHistory) > 32:
                self.__findNameHistory = self.__findNameHistory[:32]
            setFindNameHistory(self.__findNameHistory)
        self.close()

    def __enterInFilter(self):
        """Handles ENTER and RETURN keys in the find combo"""
        if self.__namesBrowser.getVisible() != 0:
            self.__namesBrowser.openCurrentItem()



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2016  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

#
# The file was taken from eric 4.4.3 and adopted for codimension.
# Original copyright:
# Copyright (c) 2007 - 2010 Detlev Offenbach <detlev@die-offenbachs.de>
#

"""Project browser with module browsing capabilities"""

import copy
from utils.pixmapcache import getIcon
from utils.globals import GlobalData
from utils.project import CodimensionProject
from .qt import QApplication
from .projectbrowsermodel import ProjectBrowserModel
from .filesbrowserbase import FilesBrowser
from .viewitems import DirectoryItemType


class ProjectBrowser(FilesBrowser):

    """Project tree browser"""

    def __init__(self, parent):

        self.__mainWindow = parent
        FilesBrowser.__init__(self, ProjectBrowserModel(self.__mainWindow),
                              True, self.__mainWindow)

        self.setWindowTitle('Project browser')
        self.setWindowIcon(getIcon('icon.png'))

        GlobalData().project.sigProjectChanged.connect(self.__onProjectChanged)
        GlobalData().project.sigFSChanged.connect(self._onFSChanged)

        # VCS status support
        GlobalData().pluginManager.sigPluginDeactivated.connect(
            self.__onPluginDeactivated)
        self.__mainWindow.vcsManager.sigVCSFileStatus.connect(
            self.__onVCSFileStatus)
        self.__mainWindow.vcsManager.sigVCSDirStatus.connect(
            self.__onVCSDirStatus)

    def __onProjectChanged(self, what):
        """Triggered when a project is changed"""
        if what == CodimensionProject.CompleteProject:
            self.model().beginResetModel()
            self.model().endResetModel()

    def reload(self):
        """Reloads the projects view"""
        self.model().sourceModel().populateModel()
        self.model().beginResetModel()
        self.model().endResetModel()
        self.layoutDisplay()

    # Arguments: plugin
    def __onPluginDeactivated(self, _):
        """Triggered when a plugin is deactivated"""
        if self.__mainWindow.vcsManager.activePluginCount() == 0:
            for treeItem in self.model().sourceModel().rootItem.childItems:
                self.__resetVCSStatus(treeItem)

    def __resetVCSStatus(self, treeItem):
        """Recursively resets the VCS status if needed"""
        if treeItem.vcsStatus:
            treeItem.vcsStatus = None
            self._signalItemUpdated(treeItem)

        for i in treeItem.childItems:
            if i.itemType == DirectoryItemType:
                self.__resetVCSStatus(i)
            else:
                if i.vcsStatus:
                    i.vcsStatus = None
                    self._signalItemUpdated(i)

    def __onVCSFileStatus(self, path, status):
        """Triggered when a status was updated"""
        for treeItem in self.model().sourceModel().rootItem.childItems:
            self._updateVCSFileStatus(treeItem, path, status)

    def _updateVCSFileStatus(self, treeItem, path, status):
        """Updates the VCS status of an item"""
        # Due to symbolic links the whole tree must be checked, so there is
        # no limit here...
        updateSent = False
        for i in treeItem.childItems:
            if i.itemType == DirectoryItemType:
                self._updateVCSFileStatus(i, path, status)
            else:
                # It is a file
                if i.getRealPath() == path:
                    if i.vcsStatus != status:
                        i.vcsStatus = copy.deepcopy(status)
                        self._signalItemUpdated(i)
                        updateSent = True
        if updateSent:
            self.__focusFlip()

    def __onVCSDirStatus(self, path, status):
        """Triggered when a status is updated"""
        for treeItem in self.model().sourceModel().rootItem.childItems:
            self._updateVCSDirStatus(treeItem, path, status)

    def _updateVCSDirStatus(self, treeItem, path, status):
        """Updates the VCS status of an item"""
        # Due to symbolic links the whole tree must be checked, so there is
        # no limit here...
        updateSent = False
        if treeItem.getRealPath() == path:
            if treeItem.vcsStatus != status:
                treeItem.vcsStatus = copy.deepcopy(status)
                self._signalItemUpdated(treeItem)
                updateSent = True

        for i in treeItem.childItems:
            if i.itemType == DirectoryItemType:
                self._updateVCSDirStatus(i, path, status)

        if updateSent:
            self.__focusFlip()

    def __focusFlip(self):
        """Dirty trick utility function"""
        # QT does not update the left part of the item where
        # a tree part is drawn. It updates only the right part of the item
        # when a signal is sent. Qt however updates the tree part when
        # focus is received. So here is focus flipping below.
        currentFocus = QApplication.focusWidget()
        if currentFocus:
            if currentFocus == self:
                editorsManager = self.__mainWindow.editorsManager()
                currentWidget = editorsManager.currentWidget()
                if currentWidget:
                    currentWidget.setFocus()
                    self.setFocus()
            else:
                self.setFocus()
                currentFocus.setFocus()

    def drawBranches(self, painter, rect, index):
        """Helps to draw the solid highlight line for the project browser.

        This part is responsible for the beginning of the background line
        till +/- icon.
        See also the ProjectBrowserModel::data(...) method which draws
        the rest of the background
        http://stackoverflow.com/questions/14255224/
        changing-the-row-background-color-of-a-qtreeview-does-not-work
        """
        if index.isValid():
            if index != self.currentIndex():
                item = self.model().item(index)
                if item.vcsStatus:
                    vcsManager = self.__mainWindow.vcsManager
                    indicator = vcsManager.getStatusIndicator(item.vcsStatus)
                    if indicator and indicator.backgroundColor:
                        painter.fillRect(rect, indicator.backgroundColor)
        FilesBrowser.drawBranches(self, painter, rect, index)



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Debugger server"""

import logging
import os.path
from ui.qt import (pyqtSignal, QTimer, QObject, QModelIndex,
                   QMessageBox, QDialog)
from utils.pixmapcache import getIcon
from .client.protocol_cdm_dbg import (METHOD_LINE, METHOD_STACK, METHOD_STEP,
                                      METHOD_THREAD_LIST, METHOD_VARIABLES,
                                      METHOD_CONTINUE, METHOD_DEBUG_STARTUP,
                                      METHOD_FORK_TO, METHOD_CLEAR_BP,
                                      METHOD_SYNTAX_ERROR, METHOD_EXCEPTION,
                                      METHOD_VARIABLE, METHOD_STEP_OVER,
                                      METHOD_BP_CONDITION_ERROR,
                                      METHOD_SET_BP, METHOD_STEP_OUT,
                                      METHOD_BP_ENABLE, METHOD_BP_IGNORE,
                                      METHOD_STEP_QUIT, METHOD_CALL_TRACE,
                                      METHOD_EXECUTE_STATEMENT,
                                      METHOD_EXEC_STATEMENT_ERROR,
                                      METHOD_EXEC_STATEMENT_OUTPUT,
                                      METHOD_SIGNAL, METHOD_THREAD_SET)
from .bputils import getBreakpointLines
from .breakpointmodel import BreakPointModel
from .watchpointmodel import WatchPointModel
from .editbreakpoint import BreakpointEditDialog


class CodimensionDebugger(QObject):

    """Debugger server implementation"""

    sigDebuggerStateChanged = pyqtSignal(int)
    sigClientLine = pyqtSignal(str, int, bool)
    sigClientException = pyqtSignal(str, str, list, bool)
    sigClientSyntaxError = pyqtSignal(str, str, str, int, int)
    sigClientStack = pyqtSignal(list)
    sigClientThreadList = pyqtSignal(int, list)
    sigClientVariables = pyqtSignal(int, list)
    sigClientVariable = pyqtSignal(int, list)
    sigClientThreadSet = pyqtSignal()
    sigClientClearBreak = pyqtSignal(str, int)
    sigClientBreakConditionError = pyqtSignal(str, int)
    sigClientCallTrace = pyqtSignal(bool, str, int, str, str, int, str)

    STATE_STOPPED = 0
    STATE_IN_CLIENT = 1
    STATE_IN_IDE = 2

    def __init__(self, mainWindow):
        QObject.__init__(self)

        # To control the user interface elements
        self.__mainWindow = mainWindow
        self.__state = self.STATE_STOPPED

        self.__stopAtFirstLine = None

        self.__procWrapper = None
        self.__procuuid = None
        self.__fileName = None
        self.__runParameters = None
        self.__debugSettings = None

        self.__breakpointModel = BreakPointModel(self)
        self.__watchpointModel = WatchPointModel(self)

        self.__breakpointModel.rowsAboutToBeRemoved.connect(
            self.__deleteBreakPoints)
        self.__breakpointModel.sigDataAboutToBeChanged.connect(
            self.__breakPointDataAboutToBeChanged)
        self.__breakpointModel.dataChanged.connect(self.__changeBreakPoints)
        self.__breakpointModel.rowsInserted.connect(self.__addBreakPoints)
        self.sigClientClearBreak.connect(self.__clientClearBreakPoint)
        self.sigClientBreakConditionError.connect(
            self.__clientBreakConditionError)

        self.__handlers = {}
        self.__initHandlers()

    def __initHandlers(self):
        """Initializes the incoming messages handlers"""
        self.__handlers = {
            METHOD_LINE: self.__handleLine,
            METHOD_STACK: self.__handleStack,
            METHOD_THREAD_LIST: self.__handleThreadList,
            METHOD_VARIABLES: self.__handleVariables,
            METHOD_DEBUG_STARTUP: self.__handleStartup,
            METHOD_FORK_TO: self.__handleForkTo,
            METHOD_CLEAR_BP: self.__handleClearBP,
            METHOD_SYNTAX_ERROR: self.__handleSyntaxError,
            METHOD_VARIABLE: self.__handleVariable,
            METHOD_BP_CONDITION_ERROR: self.__handleBPConditionError,
            METHOD_EXCEPTION: self.__handleException,
            METHOD_CALL_TRACE: self.__handleCallTrace,
            METHOD_EXEC_STATEMENT_ERROR: self.__handleExecStatementError,
            METHOD_EXEC_STATEMENT_OUTPUT: self.__handleExecuteStatementOutput,
            METHOD_SIGNAL: self.__handleSignal,
            METHOD_THREAD_SET: self.__handleThreadSet}

    def getScriptPath(self):
        """Provides the path to the debugged script"""
        return self.__fileName

    def getState(self):
        """Provides the debugger state"""
        return self.__state

    def getRunDebugParameters(self):
        """Provides the running and debugging parameters"""
        return self.__runParameters, self.__debugSettings

    def getBreakPointModel(self):
        """Provides a reference to the breakpoints model"""
        return self.__breakpointModel

    def getWatchPointModel(self):
        """Provides a reference to the watch points model"""
        return self.__watchpointModel

    def __changeDebuggerState(self, newState):
        """Changes the debugger state"""
        if newState != self.__state:
            self.__state = newState
            self.sigDebuggerStateChanged.emit(newState)

    def onDebugSessionStarted(self, procWrapper, fileName,
                              runParameters, debugSettings):
        """Starts debugging a script. Run manager informs about it."""
        if self.__state != self.STATE_STOPPED:
            raise Exception('Logic error. Debugging session started while the '
                            'previous one has not finished.')

        self.__procWrapper = procWrapper
        self.__procuuid = procWrapper.procuuid
        self.__fileName = fileName
        self.__runParameters = runParameters
        self.__debugSettings = debugSettings
        self.__stopAtFirstLine = debugSettings.stopAtFirstLine

        self.__mainWindow.switchDebugMode(True)
        self.__changeDebuggerState(self.STATE_IN_CLIENT)

    def onIncomingMessage(self, procuuid, method, params):
        """Message from the debuggee has been received"""
        if self.__procuuid == procuuid:
            try:
                self.__handlers[method](params)
            except KeyError:
                logging.error('Unhandled message received by the debugger. '
                              'Method: ' + str(method) +
                              ' Parameters: ' + repr(params))

    def __handleLine(self, params):
        """Handles METHOD_LINE"""
        stack = params['stack']
        if self.__stopAtFirstLine:
            topFrame = stack[0]
            self.sigClientLine.emit(topFrame[0], int(topFrame[1]), False)
            self.sigClientStack.emit(stack)
        else:
            self.__stopAtFirstLine = True
            QTimer.singleShot(0, self.remoteContinue)

        self.__changeDebuggerState(self.STATE_IN_IDE)

    def __handleStack(self, params):
        """Handles METHOD_STACK"""
        stack = params['stack']
        if self.__stopAtFirstLine:
            topFrame = stack[0]
            self.sigClientLine.emit(topFrame[0], int(topFrame[1]), True)
            self.sigClientStack.emit(stack)
        else:
            self.__stopAtFirstLine = True
            QTimer.singleShot(0, self.remoteContinue)

    def __handleThreadList(self, params):
        """Handles METHOD_THREAD_LIST"""
        self.sigClientThreadList.emit(params['currentID'],
                                      params['threadList'])

    def __handleVariables(self, params):
        """Handles METHOD_VARIABLES"""
        self.sigClientVariables.emit(params['scope'], params['variables'])

    def __handleStartup(self, params):
        """Handles METHOD_DEBUG_STARTUP"""
        del params  # unused argument
        self.__sendBreakpoints()
        self.__sendWatchpoints()

    def __handleForkTo(self, params):
        """Handles METHOD_FORK_TO"""
        del params  # unused argument
        self.__askForkTo()

    def __handleClearBP(self, params):
        """Handles METHOD_CLEAR_BP"""
        self.sigClientClearBreak.emit(params['filename'], params['line'])

    def __handleSyntaxError(self, params):
        """Handles METHOD_SYNTAX_ERROR"""
        self.sigClientSyntaxError.emit(self.__procuuid,
                                       params['message'], params['filename'],
                                       params['line'],
                                       params['characternumber'])

    def __handleVariable(self, params):
        """Handles METHOD_VARIABLE"""
        self.sigClientVariable.emit(params['scope'],
                                    [params['variable']] + params['variables'])

    def __handleBPConditionError(self, params):
        """Handles METHOD_BP_CONDITION_ERROR"""
        self.sigClientBreakConditionError.emit(params['filename'],
                                               params['line'])

    def __handleException(self, params):
        """Handles METHOD_EXCEPTION"""
        self.__changeDebuggerState(self.STATE_IN_IDE)
        if params:
            stack = params['stack']
            if stack:
                if stack[0] and stack[0][0] == "<string>":
                    for stackEntry in stack:
                        if stackEntry[0] == "<string>":
                            stackEntry[0] = self.__fileName
                        else:
                            break
            excType = params['type']
            isUnhandled = excType is None or \
                excType.lower().startswith('unhandled') or \
                not stack
            self.sigClientException.emit(excType, params['message'],
                                         stack, isUnhandled)
        else:
            isUnhandled = True
            self.sigClientException.emit('', '', [], True)

    def __handleCallTrace(self, params):
        """Handles METHOD_CALL_TRACE"""
        isCall = params['event'] == 'c'
        src = params['from']
        dest = params['to']
        self.sigClientCallTrace.emit(
            isCall, src['filename'], src['linenumber'], src['codename'],
            dest['filename'], dest['linenumber'], dest['codename'])

    @staticmethod
    def __handleExecStatementError(params):
        """Handles METHOD_EXEC_STATEMENT_ERROR"""
        logging.error('Execute statement error:\n' + params['text'])

    @staticmethod
    def __handleExecuteStatementOutput(params):
        """Handles METHOD_EXEC_STATEMENT_OUTPUT"""
        text = params['text']
        if text:
            logging.info('Statement execution succeeded. Output:\n' + text)
        else:
            logging.info('Statement execution succeeded. No output generated.')

    def __handleSignal(self, params):
        """Handles METHOD_SIGNAL"""
        message = params['message']
        fileName = params['filename']
        linenumber = params['linenumber']
        # funcName = params['function']
        # arguments = params['arguments']

        self.sigClientLine.emit(fileName, linenumber, False)
        logging.error('The program generated the signal "' + message + '"\n'
                      'File: ' + fileName + ' Line: ' + str(linenumber))

    def __handleThreadSet(self, params):
        """Handles METHOD_THREAD_SET"""
        del params  # unused argument
        self.sigClientThreadSet.emit()

    def onProcessFinished(self, procuuid, retCode):
        """Process finished. The retCode may indicate a disconnection."""
        del retCode     # unused argument

        if self.__procuuid == procuuid:
            self.__procWrapper = None
            self.__procuuid = None
            self.__fileName = None
            self.__runParameters = None
            self.__debugSettings = None
            self.__stopAtFirstLine = None

            self.__changeDebuggerState(self.STATE_STOPPED)
            self.__mainWindow.switchDebugMode(False)

    def __askForkTo(self):
        " Asks what to follow, a parent or a child "
        dlg = QMessageBox(QMessageBox.Question, "Client forking",
                          "Select the fork branch to follow")
        dlg.addButton(QMessageBox.Ok)
        dlg.addButton(QMessageBox.Cancel)

        btn1 = dlg.button(QMessageBox.Ok)
        btn1.setText("&Child process")
        btn1.setIcon(getIcon(''))

        btn2 = dlg.button(QMessageBox.Cancel)
        btn2.setText("&Parent process")
        btn2.setIcon(getIcon(''))

        dlg.setDefaultButton(QMessageBox.Cancel)
        res = dlg.exec_()

        if res == QMessageBox.Cancel:
            self.__sendJSONCommand(METHOD_FORK_TO, {'target': 'parent'})
        else:
            self.__sendJSONCommand(METHOD_FORK_TO, {'target': 'child'})

    def __validateBreakpoints(self):
        """Checks all the breakpoints validity and deletes invalid"""
        # It is excepted that the method is called when all the files are
        # saved, e.g. when a new debugging session is started.
        for row in range(0, self.__breakpointModel.rowCount()):
            index = self.__breakpointModel.index(row, 0, QModelIndex())
            bpoint = self.__breakpointModel.getBreakPointByIndex(index)
            fileName = bpoint.getAbsoluteFileName()
            line = bpoint.getLineNumber()

            if not os.path.exists(fileName):
                logging.warning("Breakpoint at " + fileName + ":" +
                                str(line) + " is invalid (the file "
                                "disappeared from the filesystem). "
                                "The breakpoint is deleted.")
                self.__breakpointModel.deleteBreakPointByIndex(index)
                continue

            breakableLines = getBreakpointLines(fileName, None, True)
            if breakableLines is None:
                logging.warning("Breakpoint at " + fileName + ":" +
                                str(line) + " does not point to a breakable "
                                "line (the file could not be compiled). "
                                "The breakpoint is deleted.")
                self.__breakpointModel.deleteBreakPointByIndex(index)
                continue
            if line not in breakableLines:
                logging.warning("Breakpoint at " + fileName + ":" +
                                str(line) + " does not point to a breakable "
                                "line (the file was modified). "
                                "The breakpoint is deleted.")
                self.__breakpointModel.deleteBreakPointByIndex(index)
                continue

            # The breakpoint is OK, keep it
        return

    def __sendBreakpoints(self):
        """Sends the breakpoints to the debugged program"""
        self.__validateBreakpoints()
        self.__addBreakPoints(QModelIndex(), 0,
                              self.__breakpointModel.rowCount() - 1)

    def __addBreakPoints(self, parentIndex, start, end):
        """Adds breakpoints"""
        if self.__state == self.STATE_STOPPED:
            return

        for row in range(start, end + 1):
            index = self.__breakpointModel.index(row, 0, parentIndex)
            bpoint = self.__breakpointModel.getBreakPointByIndex(index)
            fileName = bpoint.getAbsoluteFileName()
            line = bpoint.getLineNumber()
            self.remoteBreakpoint(fileName, line, True,
                                  bpoint.getCondition(),
                                  bpoint.isTemporary())
            if not bpoint.isEnabled():
                self.__remoteBreakpointEnable(fileName, line, False)
            ignoreCount = bpoint.getIgnoreCount()
            if ignoreCount > 0:
                self.__remoteBreakpointIgnore(fileName, line,
                                              ignoreCount)

    def __deleteBreakPoints(self, parentIndex, start, end):
        """Deletes breakpoints"""
        if self.__state == self.STATE_STOPPED:
            return

        for row in range(start, end + 1):
            index = self.__breakpointModel.index(row, 0, parentIndex)
            bpoint = self.__breakpointModel.getBreakPointByIndex(index)
            fileName = bpoint.getAbsoluteFileName()
            line = bpoint.getLineNumber()
            self.remoteBreakpoint(fileName, line, False)

    def __breakPointDataAboutToBeChanged(self, startIndex, endIndex):
        """Handles the sigDataAboutToBeChanged signal of the bpoint model"""
        self.__deleteBreakPoints(QModelIndex(),
                                 startIndex.row(), endIndex.row())

    def __changeBreakPoints(self, startIndex, endIndex):
        """Sets changed breakpoints"""
        self.__addBreakPoints(QModelIndex(), startIndex.row(), endIndex.row())

    def __sendWatchpoints(self):
        """Sends the watchpoints to the debugged program"""
        pass

    def __remoteBreakpointEnable(self, fileName, line, enable):
        """Sends the breakpoint enability"""
        self.__sendJSONCommand(METHOD_BP_ENABLE,
                               {'filename': fileName, 'line': line,
                                'enable': enable})

    def __remoteBreakpointIgnore(self, fileName, line, ignoreCount):
        """Sends the breakpoint ignore count"""
        self.__sendJSONCommand(METHOD_BP_IGNORE,
                               {'filename': fileName, 'line': line,
                                'count': ignoreCount})

    def __clientClearBreakPoint(self, fileName, line):
        """Handles the sigClientClearBreak signal"""
        if self.__state == self.STATE_STOPPED:
            return

        index = self.__breakpointModel.getBreakPointIndex(fileName, line)
        if index.isValid():
            self.__breakpointModel.deleteBreakPointByIndex(index)

    def __clientBreakConditionError(self, fileName, line):
        """Handles the condition error"""
        logging.error("The condition of the breakpoint at " +
                      fileName + ":" + str(line) +
                      " contains a syntax error.")
        index = self.__breakpointModel.getBreakPointIndex(fileName, line)
        if not index.isValid():
            return
        bpoint = self.__breakpointModel.getBreakPointByIndex(index)
        if not bpoint:
            return

        dlg = BreakpointEditDialog(bpoint)
        if dlg.exec_() == QDialog.Accepted:
            newBpoint = dlg.getData()
            if newBpoint == bpoint:
                return
            self.__breakpointModel.setBreakPointByIndex(index, newBpoint)

    def remoteStep(self):
        """Single step in the debugged program"""
        self.__changeDebuggerState(self.STATE_IN_CLIENT)
        self.__sendJSONCommand(METHOD_STEP, None)

    def remoteStepOver(self):
        """Step over the debugged program"""
        self.__changeDebuggerState(self.STATE_IN_CLIENT)
        self.__sendJSONCommand(METHOD_STEP_OVER, None)

    def remoteStepOut(self):
        """Step out the debugged program"""
        self.__changeDebuggerState(self.STATE_IN_CLIENT)
        self.__sendJSONCommand(METHOD_STEP_OUT, None)

    def remoteContinue(self, special=False):
        """Continues the debugged program"""
        self.__changeDebuggerState(self.STATE_IN_CLIENT)
        self.__sendJSONCommand(METHOD_CONTINUE, {'special': special})

    def remoteThreadList(self):
        """Provides the threads list"""
        self.__sendJSONCommand(METHOD_THREAD_LIST, None)

    def remoteClientVariables(self, scope, framenr=0, filters=None):
        """Provides the client variables.

        scope - 0 => local, 1 => global
        """
        if filters is None:
            filters = []
        self.__sendJSONCommand(METHOD_VARIABLES,
                               {'frameNumber': framenr,
                                'scope': scope, 'filters': filters})

    def remoteClientVariable(self, scope, var, framenr=0, filters=None):
        """Provides the client variable.

        scope - 0 => local, 1 => global
        """
        self.__sendJSONCommand(METHOD_VARIABLE,
                               {'frameNumber': framenr, 'variable': var,
                                'scope': scope, 'filters': filters})

    def remoteExecuteStatement(self, statement, framenr):
        """Executes the expression in the current context of the debuggee"""
        self.__sendJSONCommand(METHOD_EXECUTE_STATEMENT,
                               {'statement': statement,
                                'frameNumber': framenr})

    def remoteBreakpoint(self, fileName, line,
                         isSetting, condition=None, temporary=False):
        """Sets or clears a breakpoint"""
        params = {'filename': fileName, 'line': line,
                  'setBreakpoint': isSetting, 'condition': condition,
                  'temporary': temporary}
        self.__sendJSONCommand(METHOD_SET_BP, params)

    def remoteSetThread(self, tid):
        """Sets the given thread as the current"""
        self.__sendJSONCommand(METHOD_THREAD_SET, {'threadID': tid})

    def stopDebugging(self, exitCode=None):
        """Stops the debugging session"""
        if self.__procWrapper:
            if not self.__procWrapper.hasConnected():
                # The counterpart has not connected back to the IDE
                # So there is no need to send a cancel message.
                # Instead, just clean all the initialized data structures
                # in the run manager and switch to the editing mode
                self.__procWrapper.cancelPendingDebugSession()
            else:
                if exitCode is None:
                    self.__sendJSONCommand(METHOD_STEP_QUIT, None)
                else:
                    self.__sendJSONCommand(METHOD_STEP_QUIT,
                                           {'exitCode': exitCode})

    def stopCalltrace(self):
        """Sends a message to stop call tracing"""
        if self.__procWrapper:
            self.__sendJSONCommand(METHOD_CALL_TRACE, {'enable': False})

    def startCalltrace(self):
        """Sends a message to start call tracing"""
        if self.__procWrapper:
            self.__sendJSONCommand(METHOD_CALL_TRACE, {'enable': True})

    def __sendJSONCommand(self, method, params):
        """Sends a message to the debuggee"""
        if self.__procWrapper:
            self.__procWrapper.sendJSONCommand(method, params)
        else:
            raise Exception('Trying to send JSON command from the debugger '
                            'to the debugged program wneh there is no remote '
                            'process wrapper. Method: ' + str(method) +
                            'Parameters: ' + repr(params))



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2017  Sergey Satskiy sergey.satskiy@gmail.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#


"""Redirected IO console widget for running/profiling/debugging scripts"""


from ui.qt import (Qt, QSize, pyqtSignal, QToolBar, QHBoxLayout, QWidget,
                   QAction, QSizePolicy, QMenu, QToolButton)
from utils.runparams import DEBUG
from utils.settings import Settings
from utils.pixmapcache import getIcon
from .redirectedioconsole import RedirectedIOConsole


class IOConsoleWidget(QWidget):

    """IO Console widget"""

    sigSettingsUpdated = pyqtSignal()
    sigUserInput = pyqtSignal(str, str)
    sigKillIOConsoleProcess = pyqtSignal(str)
    sigCloseIOConsole = pyqtSignal(int)

    def __init__(self, procuuid, kind, parent=None):
        QWidget.__init__(self, parent)
        self.procuuid = procuuid
        self.kind = kind    # RUN/DEBUG/PROFILE

        self.__viewer = RedirectedIOConsole(self)
        self.__viewer.sigUserInput.connect(self.__onUserInput)

        self.__createLayout()

    def setFocus(self):
        """Overridden setFocus"""
        self.__viewer.setFocus()

    def __onUserInput(self, userInput):
        """The user finished input in the redirected IO console"""
        self.sigUserInput.emit(self.procuuid, userInput)
        self.__clearButton.setEnabled(True)

    def __createLayout(self):
        """Creates the toolbar and layout"""
        self.__settingsMenu = QMenu(self)
        self.__settingsMenu.aboutToShow.connect(self.__settingsAboutToShow)
        self.__wrapLongLinesAct = self.__settingsMenu.addAction(
            "Wrap long lines")
        self.__wrapLongLinesAct.setCheckable(True)
        self.__wrapLongLinesAct.triggered.connect(self.__onWrapLongLines)
        self.__showWhitespacesAct = self.__settingsMenu.addAction(
            "Show whitespaces")
        self.__showWhitespacesAct.setCheckable(True)
        self.__showWhitespacesAct.triggered.connect(self.__onShowWhitespaces)
        self.__autoscrollAct = self.__settingsMenu.addAction("Autoscroll")
        self.__autoscrollAct.setCheckable(True)
        self.__autoscrollAct.triggered.connect(self.__onAutoscroll)

        self.__settingsButton = QToolButton(self)
        self.__settingsButton.setIcon(getIcon('iosettings.png'))
        self.__settingsButton.setToolTip('View settings')
        self.__settingsButton.setPopupMode(QToolButton.InstantPopup)
        self.__settingsButton.setMenu(self.__settingsMenu)
        self.__settingsButton.setFocusPolicy(Qt.NoFocus)

        if self.kind != DEBUG:
            fixedSpacer = QWidget()
            fixedSpacer.setFixedHeight(8)

            self.__stopButton = QAction(getIcon('runconsolestop.png'),
                                        'Stop process', self)
            self.__stopButton.triggered.connect(self.stop)

        spacer = QWidget()
        spacer.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.__clearButton = QAction(getIcon('trash.png'), 'Clear', self)
        self.__clearButton.triggered.connect(self.clear)

        # The toolbar
        toolbar = QToolBar(self)
        toolbar.setOrientation(Qt.Vertical)
        toolbar.setMovable(False)
        toolbar.setAllowedAreas(Qt.RightToolBarArea)
        toolbar.setIconSize(QSize(16, 16))
        toolbar.setFixedWidth(28)
        toolbar.setContentsMargins(0, 0, 0, 0)

        toolbar.addWidget(self.__settingsButton)

        if self.kind != DEBUG:
            toolbar.addWidget(fixedSpacer)
            toolbar.addAction(self.__stopButton)

        toolbar.addWidget(spacer)
        toolbar.addAction(self.__clearButton)

        hLayout = QHBoxLayout()
        hLayout.setContentsMargins(0, 0, 0, 0)
        hLayout.setSpacing(0)
        hLayout.addWidget(toolbar)
        hLayout.addWidget(self.__viewer)

        self.setLayout(hLayout)

    def __settingsAboutToShow(self):
        """Settings menu is about to show"""
        self.__wrapLongLinesAct.setChecked(Settings()['ioconsolelinewrap'])
        self.__showWhitespacesAct.setChecked(Settings()['ioconsoleshowspaces'])
        self.__autoscrollAct.setChecked(Settings()['ioconsoleautoscroll'])

    def __onWrapLongLines(self):
        """Triggered when long lines setting is changed"""
        Settings()['ioconsolelinewrap'] = not Settings()['ioconsolelinewrap']
        self.sigSettingsUpdated.emit()

    def __onShowWhitespaces(self):
        """Triggered when show whitespaces is changed"""
        Settings()['ioconsoleshowspaces'] = \
            not Settings()['ioconsoleshowspaces']
        self.sigSettingsUpdated.emit()

    @staticmethod
    def __onAutoscroll():
        """Triggered when autoscroll is changed"""
        Settings()['ioconsoleautoscroll'] = \
            not Settings()['ioconsoleautoscroll']

    def clear(self):
        """Triggered when requested to clear the console"""
        self.__viewer.clearAll()

    def consoleSettingsUpdated(self):
        """Triggered when one of the consoles updated a common setting"""
        self.__viewer.updateSettings()

    def resizeEvent(self, event):
        """Handles the widget resize"""
        QWidget.resizeEvent(self, event)

    def writeFile(self, fileName):
        """Writes the text to a file"""
        return self.__viewer.writeFile(fileName)

    def appendIDEMessage(self, text):
        """Appends an IDE message"""
        return self.__viewer.appendIDEMessage(text)

    def appendStdoutMessage(self, _, text):
        """Appends an stdout message"""
        return self.__viewer.appendStdoutMessage(text)

    def appendStderrMessage(self, _, text):
        """Appends an stderr message"""
        return self.__viewer.appendStderrMessage(text)

    def onTextZoomChanged(self):
        """Triggered when a text zoom is changed"""
        self.__viewer.onTextZoomChanged()

    def input(self, procuuid, prompt, echo):
        """Triggered when input is requested"""
        self.__viewer.inputEcho = echo
        if prompt:
            self.appendStdoutMessage(procuuid, prompt)
        self.__clearButton.setEnabled(False)
        self.__viewer.switchMode(self.__viewer.MODE_INPUT)

    def sessionStopped(self):
        """Triggered when redirecting session is stopped"""
        self.__viewer.switchMode(self.__viewer.MODE_OUTPUT)
        self.__clearButton.setEnabled(True)

    def stop(self):
        """Triggered when the user requesed to stop the process"""
        self.sigKillIOConsoleProcess.emit(self.procuuid)

    def scriptFinished(self):
        """Triggered when the script process finished"""
        if self.kind != DEBUG:
            self.__stopButton.setEnabled(False)
        self.__viewer.switchMode(self.__viewer.MODE_OUTPUT)
        self.__clearButton.setEnabled(True)

    def onReuse(self, procuuid):
        """Triggered when the console is reused"""
        self.procuuid = procuuid
        if self.kind != DEBUG:
            self.__stopButton.setEnabled(True)



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Provides the storage for the search environment"""

import os.path
from copy import deepcopy
from .fileutils import loadJSON, saveJSON


# The 'find' item is for the search in the buffer dialogue. Each item is a
# dictionary of the following structure:
# { 'what': <string>,
#   'case': <bool>, 'word': <bool>, 'regexp': <bool> }
#
# The 'replace' item is for the replace in the buffer dialogue. Each item is a
# dictionary of the following structure:
# { 'what': <string>, 'with': <string>,
#   'case': <bool>, 'word': <bool>, 'regexp': <bool> }
#
# The 'findinfiles' item is for a modal find in files dialogue box. Each item
# is a dictionary of the following structure:
# { 'what': <string>,
#   'case': <bool>, 'word': <bool>, 'regexp': <bool>,
#   'inproject': <bool>, 'inopened': <bool>, 'indir': <bool>, 'dir': <string>,
#   'filter': <string> }
_DEFAULT_SEARCH_HISTORY = {
    'class': [],        # [term, ...]
    'function': [],     # [term, ...]
    'global': [],       # [term, ...]
    'findname': [],     # [term, ...]
    'findfile': [],     # [term, ...]
    'find': [],         # [ {'term': , 'replace': ,
                        #    'cbCase': , 'cbWord': , 'cbRegexp': }, ... ]
    'findinfiles': []}  # [ {'term': , 'dir': , 'filters': ,
                        #    'cbCase': , 'cbWord': , 'cbRegexp': ,
                        #    'rbProject': , 'rbOpen': , 'rbDir': }, ... ]

class SearchEnvironment:

    """Loads/stores/saves the search environment"""

    def __init__(self):
        self.__props = deepcopy(_DEFAULT_SEARCH_HISTORY)
        self.__seFileName = None

        # Default. Could be updated later
        self.__limit = 32

    def reset(self):
        """Un-binds from the file system"""
        self.__props = deepcopy(_DEFAULT_SEARCH_HISTORY)
        self.__seFileName = None

    def setup(self, dirName):
        """Binds the parameters to a disk file"""
        # Just in case - flush the previous data if they were bound
        SearchEnvironment.save(self)

        dirName = os.path.realpath(dirName)
        if not dirName.endswith(os.path.sep):
            dirName += os.path.sep
        if not os.path.isdir(dirName):
            raise Exception('Directory name is expected for the search '
                            'environment. The given ' + dirName + ' is not.')

        self.__seFileName = dirName + "searchenv.json"
        if os.path.exists(self.__seFileName):
            SearchEnvironment.load(self)

    def load(self):
        """Loads the saved search environment"""
        if self.__seFileName:
            default = deepcopy(_DEFAULT_SEARCH_HISTORY)
            self.__props = loadJSON(self.__seFileName, 'search environment',
                                    default)

    def save(self):
        """Saves the search environment into a file"""
        if self.__seFileName:
            saveJSON(self.__seFileName, self.__props, 'search environment')

    def __addToContainer(self, element, item):
        """Common implementation of adding a search item"""
        if item in self.__props[element]:
            self.__props[element].remove(item)
        self.__props[element].insert(0, item)
        if len(self.__props[element]) > self.__limit:
            self.__props[element] = self.__props[element][0:self.__limit]
        SearchEnvironment.save(self)

    def __setContainer(self, item, history):
        """Generic container setter which respects the limit"""
        if len(history) > self.__limit:
            self.__props[item] = history[0:self.__limit]
        else:
            self.__props[item] = history
        SearchEnvironment.save(self)

    def setLimit(self, newLimit):
        """Sets the new limit"""
        self.__limit = newLimit

    @property
    def findClassHistory(self):
        """Provides the find class history"""
        return self.__props['class']

    @findClassHistory.setter
    def findClassHistory(self, history):
        self.__setContainer('class', history)

    def addToFindClassHistory(self, item):
        """Adds an item to the class history"""
        self.__addToContainer('class', item)

    @property
    def findFunctionHistory(self):
        """Provides the find function history"""
        return self.__props['function']

    @findFunctionHistory.setter
    def findFunctionHistory(self, history):
        self.__setContainer('function', history)

    def addToFindFunctionHistory(self, item):
        """Adds an item to the function history"""
        self.__addToContainer('function', item)

    @property
    def findGlobalHistory(self):
        """Provides the find global history"""
        return self.__props['global']

    @findGlobalHistory.setter
    def findGlobalHistory(self, history):
        self.__setContainer('global', history)

    def addToFindGlobalHistory(self, item):
        """Adds an item to the global history"""
        self.__addToContainer('global', item)

    @property
    def findNameHistory(self):
        """Provides the find name history"""
        return self.__props['findname']

    @findNameHistory.setter
    def findNameHistory(self, history):
        self.__setContainer('findname', history)

    def addToFindNameHistory(self, item):
        """Adds an item to the name history"""
        self.__addToContainer('findname', item)

    @property
    def findFileHistory(self):
        """Provides the find file history"""
        return self.__props['findfile']

    @findFileHistory.setter
    def findFileHistory(self, history):
        self.__setContainer('findfile', history)

    def addToFindFileHistory(self, item):
        """Adds an item to the file history"""
        self.__addToContainer('findfile', item)

    @property
    def findHistory(self):
        """Provides the find history"""
        return self.__props['find']

    @findHistory.setter
    def findHistory(self, history):
        self.__setContainer('find', history)

    def addToFindHistory(self, item):
        """Adds an item to the file history"""
        self.__addToContainer('find', item)

    @property
    def findInFilesHistory(self):
        """Provides the find in files history"""
        return self.__props['findinfiles']

    @findInFilesHistory.setter
    def findInFilesHistory(self, history):
        self.__setContainer('findinfiles', history)

    def addToFindInFilesHistory(self, item):
        """Adds an item to the file history"""
        self.__addToContainer('findinfiles', item)



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#


"""A few constants which do not depend on other project files"""

# Default encoding for the cases when:
# - the encoding could not be detected
# - replaces ascii to be on the safe side
DEFAULT_ENCODING = 'utf-8'

# File encoding used for various settings and project files
SETTINGS_ENCODING = 'utf-8'

# Directory to store Codimension settings and projects
CONFIG_DIR = '.codimension3'



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2012-2018  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#


"""not used code analysis using vulture"""


import os
import os.path
import logging
import tempfile
from subprocess import Popen, PIPE
from ui.qt import (QCursor, Qt, QTimer, QDialog, QDialogButtonBox, QVBoxLayout,
                   QLabel, QApplication)
from ui.findinfiles import ItemToSearchIn, getSearchItemIndex
from utils.globals import GlobalData
from utils.config import DEFAULT_ENCODING



class NotUsedAnalysisProgress(QDialog):

    """Progress of the not used analysis"""

    def __init__(self, path, parent=None):
        QDialog.__init__(self, parent)

        path = os.path.abspath(path)
        if not os.path.exists(path):
            raise Exception('Dead code analysis path must exist. '
                            'The provide path "' + path + '" does not.')

        self.__path = path

        self.__cancelRequest = False
        self.__inProgress = False

        self.__infoLabel = None
        self.__foundLabel = None
        self.__found = 0        # Number of found

        self.__createLayout()
        if os.path.isdir(path):
            self.setWindowTitle('Dead code analysis for all project files')
        else:
            self.setWindowTitle('Dead code analysis for ' +
                                os.path.basename(path))
        self.__updateFoundLabel()
        QTimer.singleShot(0, self.__process)

    def keyPressEvent(self, event):
        """Processes the ESC key specifically"""
        if event.key() == Qt.Key_Escape:
            self.__onClose()
        else:
            QDialog.keyPressEvent(self, event)

    def __updateFoundLabel(self):
        """Updates the found label"""
        text = "Found: " + str(self.__found) + " candidate"
        if self.__found != 1:
            text += "s"
        self.__foundLabel.setText(text)

    def __createLayout(self):
        """Creates the dialog layout"""
        self.resize(450, 20)
        self.setSizeGripEnabled(True)

        verticalLayout = QVBoxLayout(self)

        # Note label
        noteLabel = QLabel("<b>Note</b>: the analysis is "
                           "suggestive and not precise. "
                           "Use the results with caution.\n", self)
        verticalLayout.addWidget(noteLabel)

        # Info label
        self.__infoLabel = QLabel(self)
        verticalLayout.addWidget(self.__infoLabel)

        # Found label
        self.__foundLabel = QLabel(self)
        verticalLayout.addWidget(self.__foundLabel)

        # Buttons
        buttonBox = QDialogButtonBox(self)
        buttonBox.setOrientation(Qt.Horizontal)
        buttonBox.setStandardButtons(QDialogButtonBox.Close)
        verticalLayout.addWidget(buttonBox)

        buttonBox.rejected.connect(self.__onClose)

    def __onClose(self):
        """triggered when the close button is clicked"""
        self.__cancelRequest = True
        if not self.__inProgress:
            self.close()

    def __run(self):
        """Runs vulture"""
        errTmp = tempfile.mkstemp()
        errStream = os.fdopen(errTmp[0])
        process = Popen(['vulture', self.__path],
                        stdin=PIPE,
                        stdout=PIPE, stderr=errStream)
        process.stdin.close()
        processStdout = process.stdout.read()
        process.stdout.close()
        errStream.seek(0)
        err = errStream.read()
        errStream.close()
        process.wait()
        try:
            os.unlink(errTmp[1])
        except:
            pass
        return processStdout.decode(DEFAULT_ENCODING), err.strip()

    def __process(self):
        """Analysis process"""
        self.__inProgress = True
        mainWindow = GlobalData().mainWindow

        QApplication.setOverrideCursor(QCursor(Qt.WaitCursor))

        # return code gives really nothing. So the error in running the utility
        # is detected by the stderr content.
        # Also, there could be a mix of messages for a project. Some files
        # could have syntax errors - there will be messages on stderr. The
        # other files are fine so there will be messages on stdout
        stdout, stderr = self.__run()
        candidates = []
        for line in stdout.splitlines():
            line = line.strip()
            if line:
                # Line is like file.py:2: unused variable 'a' (60% confidence)
                try:
                    startIndex = line.find(':')
                    if startIndex < 0:
                        continue
                    endIndex = line.find(':', startIndex + 1)
                    if endIndex < 0:
                        continue
                    fileName = line[:startIndex]
                    startIndex = line.find(':')
                    if startIndex < 0:
                        continue
                    endIndex = line.find(':', startIndex + 1)
                    if endIndex < 0:
                        continue
                    fileName = os.path.abspath(line[:startIndex])
                    lineno = int(line[startIndex + 1:endIndex])
                    message = line[endIndex + 1:].strip()
                except:
                    continue

                index = getSearchItemIndex(candidates, fileName)
                if index < 0:
                    widget = mainWindow.getWidgetForFileName(fileName)
                    if widget is None:
                        uuid = ''
                    else:
                        uuid = widget.getUUID()
                    newItem = ItemToSearchIn(fileName, uuid)
                    candidates.append(newItem)
                    index = len(candidates) - 1
                candidates[index].addMatch('', lineno, message)

                self.__found += 1
                self.__updateFoundLabel()
                QApplication.processEvents()

        if self.__found == 0:
            if stderr:
                logging.error('Error running vulture for ' + self.__path +
                              ':\n' + stderr)
            else:
                logging.info('No unused candidates found')
        else:
            mainWindow.displayFindInFiles('', candidates)

        QApplication.restoreOverrideCursor()
        self.__infoLabel.setText('Done')
        self.__inProgress = False

        self.accept()



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy sergey.satskiy@gmail.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Redirected IO console implementation"""

import logging
from qutepart import Qutepart
from ui.qt import (Qt, QEvent, pyqtSignal, QFontMetrics, QMenu,
                   QApplication, QTextOption)
from utils.pixmapcache import getIcon
from utils.globals import GlobalData
from utils.settings import Settings
from .qpartwrap import QutepartWrapper
from .redirectedmsg import IOConsoleMessages, IOConsoleMsg
from .redirectediomargin import CDMRedirectedIOMargin


CTRL_SHIFT = int(Qt.ShiftModifier | Qt.ControlModifier)
SHIFT = int(Qt.ShiftModifier)
CTRL = int(Qt.ControlModifier)
CTRL_KEYPAD = int(Qt.KeypadModifier | Qt.ControlModifier)
NO_MODIFIER = int(Qt.NoModifier)


class RedirectedIOConsole(QutepartWrapper):

    """Widget which implements the redirected IO console"""

    sigUserInput = pyqtSignal(str)

    MODE_OUTPUT = 0
    MODE_INPUT = 1

    def __init__(self, parent):
        QutepartWrapper.__init__(self, parent)

        self.setAttribute(Qt.WA_KeyCompression)

        self.mode = self.MODE_OUTPUT
        self.lastOutputPos = None
        self.inputEcho = True
        self.inputBuffer = ""
        self.__messages = IOConsoleMessages()

        self.__initGeneralSettings()
        self.__initMargins()
        self._initContextMenu()
        self.onTextZoomChanged()

        self.__hotKeys = {}
        self.__initHotKeys()
        self.installEventFilter(self)

        self.cursorPositionChanged.connect(self.setCursorStyle)

    def __initHotKeys(self):
        """Initializes a map for the hot keys event filter"""
        self.autoIndentLineAction.setShortcut('Ctrl+Shift+I')
        self.invokeCompletionAction.setEnabled(False)
        self.__hotKeys = {
            CTRL_SHIFT: {Qt.Key_C: self.onCtrlShiftC},
            SHIFT: {Qt.Key_End: self.onShiftEnd,
                    Qt.Key_Home: self.onShiftHome,
                    Qt.Key_Insert: self.onPasteText,
                    Qt.Key_Delete: self.onShiftDel},
            CTRL: {Qt.Key_V: self.onPasteText,
                   Qt.Key_X: self.onShiftDel,
                   Qt.Key_C: self.onCtrlC,
                   Qt.Key_Insert: self.onCtrlC,
                   Qt.Key_Minus: Settings().onTextZoomOut,
                   Qt.Key_Equal: Settings().onTextZoomIn,
                   Qt.Key_0: Settings().onTextZoomReset,
                   Qt.Key_Home: self.onFirstChar,
                   Qt.Key_End: self.onLastChar},
            CTRL_KEYPAD: {Qt.Key_Minus: Settings().onTextZoomOut,
                          Qt.Key_Plus: Settings().onTextZoomIn,
                          Qt.Key_0: Settings().onTextZoomReset},
            NO_MODIFIER: {Qt.Key_Home: self.onHome,
                          Qt.Key_End: self.moveToLineEnd}}

    def eventFilter(self, _, event):
        """Event filter to catch shortcuts on UBUNTU"""
        if event.type() == QEvent.KeyPress:
            key = event.key()
            modifiers = int(event.modifiers())
            try:
                if modifiers in self.__hotKeys:
                    if key in self.__hotKeys[modifiers]:
                        self.__hotKeys[modifiers][key]()
                        return True

                if modifiers == NO_MODIFIER:
                    if key in [Qt.Key_Delete, Qt.Key_Backspace]:
                        if not self.__isCutDelAvailable():
                            return True
            except Exception as exc:
                logging.warning(str(exc))
        return False

    def keyPressEvent(self, event):
        """Triggered when a key is pressed"""
        key = event.key()
        if key == Qt.Key_Escape:
            self.clearSearchIndicators()
            return

        if self.mode == self.MODE_OUTPUT:
            if key in [Qt.Key_Left, Qt.Key_Up, Qt.Key_Right, Qt.Key_Down,
                       Qt.Key_PageUp, Qt.Key_PageDown]:
                QutepartWrapper.keyPressEvent(self, event)
            return

        # It is an input mode
        txt = event.text()
        if len(txt) and txt >= ' ':
            # Printable character
            if self.absCursorPosition < self.lastOutputPos:
                # Out of the input zone
                return

            if self.inputEcho:
                QutepartWrapper.keyPressEvent(self, event)
            else:
                self.inputBuffer += txt
            return

        # Non-printable character or some other key
        if key == Qt.Key_Enter or key == Qt.Key_Return:
            userInput = self.__getUserInput()
            self.switchMode(self.MODE_OUTPUT)
            timestampLine, _ = self.getEndPosition()
            self.append('\n')
            self.clearUndoRedoHistory()
            line, pos = self.getEndPosition()
            self.cursorPosition = line, pos
            self.ensureLineOnScreen(line)
            msg = IOConsoleMsg(IOConsoleMsg.STDIN_MESSAGE,
                               userInput + '\n')
            self.__messages.append(msg)

            # margin data
            timestamp = msg.getTimestamp()
            margin = self.getMargin('cdm_redirected_io_margin')
            margin.addData(timestampLine + 1, timestamp,
                           timestamp, IOConsoleMsg.STDIN_MESSAGE)

            self.sigUserInput.emit(userInput)
            return

        if key == Qt.Key_Backspace:
            if self.absCursorPosition == self.lastOutputPos:
                if not self.inputEcho:
                    self.inputBuffer = self.inputBuffer[:-1]
                return

        QutepartWrapper.keyPressEvent(self, event)

    def mouseDoubleClickEvent(self, event):
        """Disable search highlight on double click"""
        Qutepart.mouseDoubleClickEvent(self, event)

    def onPasteText(self):
        """Triggered when insert is requested"""
        if self.mode == self.MODE_OUTPUT:
            return
        if self.absCursorPosition < self.lastOutputPos:
            return

        # Check what is in the buffer
        text = QApplication.clipboard().text()
        if '\n' in text or '\r' in text:
            return

        if not self.inputEcho:
            self.inputBuffer += text
            return

        self.paste()

    def __getUserInput(self):
        """Provides the collected user input"""
        if self.mode != self.MODE_INPUT:
            return ''
        if self.inputEcho:
            _, endPos = self.getEndPosition()
            _, beginPos = self.mapToLineCol(self.lastOutputPos)
            return self.lines[-1][beginPos:endPos]
        value = self.inputBuffer
        self.inputBuffer = ""
        return value

    def __initGeneralSettings(self):
        """Sets some generic look and feel"""
        skin = GlobalData().skin

        self.updateSettings()

        self.setPaper(skin['ioconsolePaper'])
        self.setColor(skin['ioconsoleColor'])

        self.currentLineColor = None
        self.lineLengthEdge = None
        self.setCursorStyle()

    def updateSettings(self):
        """Updates the IO console settings"""
        if Settings()['ioconsolelinewrap']:
            self.setWordWrapMode(QTextOption.WrapAnywhere)
        else:
            self.setWordWrapMode(QTextOption.NoWrap)

        self.drawAnyWhitespace = Settings()['ioconsoleshowspaces']
        self.drawIncorrectIndentation = Settings()['ioconsoleshowspaces']

    def setCursorStyle(self):
        """Sets the cursor style depending on the mode and the cursor pos"""
        if self.mode == self.MODE_OUTPUT:
            if self.cursorWidth() != 1:
                self.setCursorWidth(1)
        else:
            if self.absCursorPosition >= self.lastOutputPos:
                if self.cursorWidth() == 1:
                    fontMetrics = QFontMetrics(self.font(), self)
                    self.setCursorWidth(fontMetrics.width('W'))
                    self.update()
            else:
                if self.cursorWidth() != 1:
                    self.setCursorWidth(1)
                    self.update()

    def switchMode(self, newMode):
        """Switches between input/output mode"""
        self.mode = newMode
        if self.mode == self.MODE_OUTPUT:
            self.lastOutputPos = None
            self.inputEcho = True
            self.inputBuffer = ""
        else:
            line, pos = self.getEndPosition()
            self.cursorPosition = line, pos
            self.lastOutputPos = self.absCursorPosition
            self.ensureLineOnScreen(line)
        self.setCursorStyle()

    def __initMargins(self):
        """Initializes the IO console margins"""
        # The supported margins: timestamp
        self.addMargin(CDMRedirectedIOMargin(self))

    def _initContextMenu(self):
        """Called to initialize a context menu"""
        self._menu = QMenu(self)
        self.__menuUndo = self._menu.addAction(
            getIcon('undo.png'), '&Undo', self.onUndo, "Ctrl+Z")
        self.__menuRedo = self._menu.addAction(
            getIcon('redo.png'), '&Redo', self.onRedo, "Ctrl+Y")
        self._menu.addSeparator()
        self.__menuCut = self._menu.addAction(
            getIcon('cutmenu.png'), 'Cu&t', self.onShiftDel, "Ctrl+X")
        self.__menuCopy = self._menu.addAction(
            getIcon('copymenu.png'), '&Copy', self.onCtrlC, "Ctrl+C")
        self.__menucopyTimestamp = self._menu.addAction(
            getIcon('copymenu.png'), '&Copy all with timestamps',
            self.onCtrlShiftC, "Ctrl+Shift+C")
        self.__menuPaste = self._menu.addAction(
            getIcon('pastemenu.png'), '&Paste', self.onPasteText, "Ctrl+V")
        self.__menuSelectAll = self._menu.addAction(
            getIcon('selectallmenu.png'), 'Select &all',
            self.selectAll, "Ctrl+A")
        self._menu.addSeparator()
        self.__menuOpenAsFile = self._menu.addAction(
            getIcon('filemenu.png'), 'O&pen as file', self.openAsFile)
        self.__menuDownloadAndShow = self._menu.addAction(
            getIcon('filemenu.png'), 'Do&wnload and show',
            self.downloadAndShow)
        self.__menuOpenInBrowser = self._menu.addAction(
            getIcon('homepagemenu.png'), 'Open in browser', self.openInBrowser)
        self._menu.addSeparator()

        self._menu.aboutToShow.connect(self._contextMenuAboutToShow)
        self._menu.aboutToHide.connect(self._contextMenuAboutToHide)

    def contextMenuEvent(self, event):
        """Called just before showing a context menu"""
        event.accept()
        self._menu.popup(event.globalPos())

    def _contextMenuAboutToShow(self):
        """IO Console context menu is about to show"""
        self.__menuUndo.setEnabled(self.document().isUndoAvailable())
        self.__menuRedo.setEnabled(self.document().isRedoAvailable())

        pasteText = QApplication.clipboard().text()
        pasteEnable = pasteText != "" and \
                      '\n' not in pasteText and \
                      '\r' not in pasteText and \
                      self.mode != self.MODE_OUTPUT
        if pasteEnable:
            if self.absCursorPosition < self.lastOutputPos:
                pasteEnable = False

        # Need to make decision about menu items for modifying the input
        self.__menuCut.setEnabled(self.__isCutDelAvailable())
        self.__menuCopy.setEnabled(self.__messages.size > 0)
        self.__menucopyTimestamp.setEnabled(self.__messages.size > 0)
        self.__menuPaste.setEnabled(pasteEnable)
        self.__menuSelectAll.setEnabled(self.__messages.size > 0)

        self.__menuOpenAsFile.setEnabled(self.openAsFileAvailable())
        self.__menuDownloadAndShow.setEnabled(
            self.downloadAndShowAvailable())
        self.__menuOpenInBrowser.setEnabled(
            self.downloadAndShowAvailable())

    def _contextMenuAboutToHide(self):
        """IO console context menu is about to hide"""
        self.__menuUndo.setEnabled(True)
        self.__menuRedo.setEnabled(True)
        self.__menuCut.setEnabled(True)
        self.__menuCopy.setEnabled(True)
        self.__menucopyTimestamp.setEnabled(True)
        self.__menuPaste.setEnabled(True)
        self.__menuSelectAll.setEnabled(True)
        self.__menuOpenAsFile.setEnabled(True)
        self.__menuDownloadAndShow.setEnabled(True)
        self.__menuOpenInBrowser.setEnabled(True)

    def __isCutDelAvailable(self):
        """Returns True if cutting or deletion is possible"""
        if self.mode == self.MODE_OUTPUT:
            return False
        if self.selectedText:
            startPosition, cursorPosition = self.absSelectedPosition
            minPos = min(startPosition, cursorPosition)
            return minPos >= self.lastOutputPos
        return self.absCursorPosition > self.lastOutputPos

    def onShiftDel(self):
        """Deletes the selected text"""
        if self.selectedText:
            if self.__isCutDelAvailable():
                self.cut()
            return True
        return True

    def onUndo(self):
        """undo implementation"""
        if self.document().isUndoAvailable():
            self.undo()

    def onRedo(self):
        """redo implementation"""
        if self.document().isRedoAvailable():
            self.redo()

    def onCtrlShiftC(self):
        """Copy all with timestamps"""
        QApplication.clipboard().setText(
            self.__messages.renderWithTimestamps())

    def appendIDEMessage(self, text):
        """Appends an IDE message"""
        msg = IOConsoleMsg(IOConsoleMsg.IDE_MESSAGE, text)
        self.__appendMessage(msg)
        return msg

    def appendStdoutMessage(self, text):
        """Appends an stdout message"""
        msg = IOConsoleMsg(IOConsoleMsg.STDOUT_MESSAGE, text)
        self.__appendMessage(msg)
        return msg

    def appendStderrMessage(self, text):
        """Appends an stderr message"""
        msg = IOConsoleMsg(IOConsoleMsg.STDERR_MESSAGE, text)
        self.__appendMessage(msg)
        return msg

    def __appendMessage(self, message):
        """Appends a new message to the console"""
        if not self.__messages.append(message):
            # There was no trimming of the message list
            self.__renderMessage(message)
        else:
            # Some messages were stripped
            self.renderContent()

    def renderContent(self):
        """Regenerates the viewer content"""
        self.clear()
        self.getMargin('cdm_redirected_io_margin').clear()
        for msg in self.__messages.msgs:
            self.__renderMessage(msg)

    def __renderMessage(self, msg):
        """Adds a single message"""
        margin = self.getMargin('cdm_redirected_io_margin')
        timestamp = msg.getTimestamp()
        if msg.msgType == IOConsoleMsg.IDE_MESSAGE:
            line, pos = self.getEndPosition()

            txt = msg.msgText
            startMarkLine = line
            if pos != 0:
                txt = '\n' + txt
                startMarkLine += 1

            self.append(txt)

            line, _ = self.getEndPosition()
            for lineNo in range(startMarkLine, line + 1):
                margin.addData(lineNo + 1, timestamp,
                               timestamp, IOConsoleMsg.IDE_MESSAGE)
        else:
            line, pos = self.getEndPosition()
            txt = msg.msgText

            startTimestampLine = line
            if pos != 0:
                lastMsgType = margin.getLineMessageType(line + 1)
                if lastMsgType == IOConsoleMsg.IDE_MESSAGE:
                    txt = '\n' + txt
                    startTimestampLine = line + 1

            self.append(txt)
            endTimestampLine, pos = self.getEndPosition()
            if pos == 0:
                endTimestampLine -= 1

            for lineNo in range(startTimestampLine, endTimestampLine + 1):
                margin.addData(lineNo + 1, timestamp, timestamp,
                               msg.msgType)

        self.clearUndoRedoHistory()
        if Settings()['ioconsoleautoscroll']:
            line, pos = self.getEndPosition()
            self.gotoLine(line + 1, pos + 1)

    def clearData(self):
        """Clears the collected data"""
        self.__messages.clear()
        self.getMargin('cdm_redirected_io_margin').clear()

    def clearAll(self):
        """Clears both data and visible content"""
        self.clearData()
        self.clear()
        self.clearUndoRedoHistory()



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2016  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Provides the storage for the file system environment"""

import os.path
from copy import deepcopy
from .fileutils import loadJSON, saveJSON


# toplevel dirs: those which are added to the file system browser
# filebrowserexpandeddirs: dirs in the project browser which were expanded when
# the user closed the project
_DEFAULT_FS_PROPS = {'tabs': [],                       # [bool: active,
                                                       #  string: path, ...]
                     'recent': [],                     # [path, ...]
                     'fsbrowserexpandeddirs': [],      # [path, ...]
                     'topleveldirs': []}               # [path, ...]


class FileSystemEnvironment:

    """Loads/stores/saves the fs related environment"""

    def __init__(self):
        self.__props = deepcopy(_DEFAULT_FS_PROPS)
        self.__fseFileName = None

        # Default. Could be updated later.
        self.__limit = 32

    def reset(self):
        """Resets the binding to the file system"""
        self.__props = deepcopy(_DEFAULT_FS_PROPS)
        self.__fseFileName = None

    def setup(self, dirName):
        """Binds the parameters to a disk file"""
        # Just in case - flush the previous data if they were bound
        FileSystemEnvironment.save(self)

        dirName = os.path.realpath(dirName)
        if not dirName.endswith(os.path.sep):
            dirName += os.path.sep
        if not os.path.isdir(dirName):
            raise Exception('Directory name is expected for the file system '
                            'environment. The given ' + dirName + ' is not.')

        self.__fseFileName = dirName + "fsenv.json"
        if os.path.exists(self.__fseFileName):
            FileSystemEnvironment.load(self)

    def load(self):
        """Loads the saved file system environment"""
        if self.__fseFileName:
            default = deepcopy(_DEFAULT_FS_PROPS)
            self.__props = loadJSON(self.__fseFileName,
                                    'file system environment', default)

    def save(self):
        """Saves the file system environment into a file"""
        if self.__fseFileName:
            saveJSON(self.__fseFileName, self.__props,
                     'file system environment')

    def setLimit(self, newLimit):
        """Sets the new limit to the number of entries"""
        self.__limit = newLimit

    @property
    def tabStatus(self):
        """Provides the opened tabs status"""
        return self.__props['tabs']

    @tabStatus.setter
    def tabStatus(self, newStatus):
        self.__props['tabs'] = newStatus
        FileSystemEnvironment.save(self)

    @property
    def recentFiles(self):
        """Provides the recently used files list"""
        return self.__props['recent']

    @recentFiles.setter
    def recentFiles(self, files):
        self.__props['recent'] = files
        FileSystemEnvironment.save(self)

    def addRecentFile(self, path):
        """Adds a single recent file. True if a new file was inserted."""
        if path in self.__props['recent']:
            self.__props['recent'].remove(path)
            self.__props['recent'].insert(0, path)
            FileSystemEnvironment.save(self)
            return False
        self.__props['recent'].insert(0, path)
        if len(self.__props['recent']) > self.__limit:
            self.__props['recent'] = self.__props['recent'][0:self.__limit]
        FileSystemEnvironment.save(self)
        return True

    def removeRecentFile(self, path):
        """Removes a single recent file"""
        if path in self.__props['recent']:
            self.__props['recent'].remove(path)
            FileSystemEnvironment.save(self)

    @property
    def fsBrowserExpandedDirs(self):
        """Provides the file system browser expanded dirs"""
        return self.__props['fsbrowserexpandeddirs']

    @fsBrowserExpandedDirs.setter
    def fsBrowserExpandedDirs(self, newDirs):
        self.__props['fsbrowserexpandeddirs'] = newDirs
        FileSystemEnvironment.save(self)

    @property
    def topLevelDirs(self):
        """Provides a list of dirs in the FS browser"""
        return self.__props['topleveldirs']

    @topLevelDirs.setter
    def topLevelDirs(self, newDirs):
        self.__props['topleveldirs'] = newDirs
        FileSystemEnvironment.save(self)

    def addTopLevelDir(self, path):
        """Adds a top level dir"""
        if not path.endswith(os.path.sep):
            path += os.path.sep
        if path not in self.__props['topleveldirs']:
            self.__props['topleveldirs'].append(path)
            FileSystemEnvironment.save(self)

    def removeTopLevelDir(self, path):
        """Removes a top level dir"""
        if not path.endswith(os.path.sep):
            path += os.path.sep
        if path in self.__props['topleveldirs']:
            self.__props['topleveldirs'].remove(path)
            FileSystemEnvironment.save(self)



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2016  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Skins support"""

# pylint: disable=C0305
# pylint: disable=W0603
# pylint: disable=W0702
# pylint: disable=W0703

import sys
import logging
import os.path
import json
from copy import deepcopy
from ui.qt import QColor, QFont
from .colorfont import (buildFont, buildColor,
                        colorFontToJSON, colorFontFromJSON)
from .fileutils import saveToFile, getFileContent
from .config import DEFAULT_ENCODING
from .settings import SETTINGS_DIR


PACKAGE_SKIN_DIR = os.path.dirname(os.path.realpath(sys.argv[0])) + \
                   os.path.sep + 'skins' + os.path.sep
USER_SKIN_DIR = SETTINGS_DIR + 'skins' + os.path.sep
OVERRIDE_FILE = 'override.json'
SAMPLE_SKIN = 'sample'

isMac = sys.platform.lower() == 'darwin'

_DEFAULT_SKIN_SETTINGS = {
    'name': 'default',
    'dark': False,
    'marginPaper': QColor(228, 228, 228, 255),
    'marginPaperDebug': QColor(255, 228, 228, 255),
    'marginColor': QColor(128, 128, 128, 255),
    'marginColorDebug': QColor(128, 128, 128, 255),

    'flakesMarginPaper': QColor(208, 208, 208, 255),
    'flakesMarginPaperDebug':  QColor(255, 228, 228, 255),

    'bpointsMarginPaper': QColor(192, 192, 192, 255),

    'findNoMatchPaper': QColor(255, 193, 204, 100),
    'findMatchPaper': QColor(164, 198, 57, 100),
    'findInvalidPaper': QColor(228, 208, 10, 100),

    # 'revisionMarginPaper': QColor(228, 228, 228, 255),
    # 'revisionMarginColor': QColor(0, 128, 0, 255),
    # 'revisionAlterPaper': QColor(238, 240, 241, 255),

    'lineNumFont': buildFont('Courier,12,-1,5,50,0,0,0,0,0') if isMac else
                   buildFont('Monospace,12,-1,5,50,0,0,0,0,0'),
    'searchMarkColor': QColor(0, 255, 0, 255),
    'searchMarkPaper': QColor(255, 0, 255, 255),
    'matchMarkColor': QColor(0, 0, 255, 255),
    'matchMarkPaper': QColor(255, 255, 0, 255),
    'nolexerPaper': QColor(255, 255, 255, 255),
    'nolexerColor': QColor(0, 0, 0, 255),
    'monoFont': buildFont('Courier,12,-1,5,50,0,0,0,0,0') if isMac else
                buildFont('Monospace,12,-1,5,50,0,0,0,0,0'),
    'currentLinePaper': QColor(232, 232, 255, 255),
    'edgeColor': QColor(127, 127, 127, 128),
    'matchedBracePaper': QColor(132, 117, 245, 255),
    'matchedBraceColor': QColor(255, 255, 255, 255),
    'unmatchedBracePaper': QColor(250, 89, 68, 255),
    'unmatchedBraceColor': QColor(0, 0, 255, 255),
    'indentGuidePaper': QColor(230, 230, 230, 255),
    'indentGuideColor': QColor(127, 127, 127, 255),
    'debugCurrentLineMarkerPaper': QColor(255, 255, 127, 255),
    'debugCurrentLineMarkerColor': QColor(0, 0, 255, 255),
    'debugExcptLineMarkerPaper': QColor(255, 64, 64, 255),
    'debugExcptLineMarkerColor': QColor(255, 255, 127, 255),
    'calltipPaper': QColor(220, 255, 220, 255),
    'calltipColor': QColor(0, 0, 0, 255),
    'calltipHighColor': QColor(250, 89, 68, 255),
    'outdatedOutlineColor': QColor(255, 154, 154, 255),

    # 'diffchanged2Color': QColor(0, 0, 0, 255),
    # 'diffchanged2Paper': QColor(247, 254, 0, 255),
    # 'diffponctColor': QColor(166, 72, 72, 255),
    # 'difflineColor': QColor(102, 102, 102, 255),
    # 'diffthColor': QColor(255, 255, 255, 255),
    # 'diffthPaper': QColor(102, 102, 102, 255),
    # 'diffaddedPaper': QColor(197, 250, 175, 255),
    # 'diffchangedColor': QColor(102, 102, 102, 255),
    # 'diffchangedPaper': QColor(244, 255, 221, 255),
    # 'diffdeletedPaper': QColor(255, 204, 204, 255),
    # 'diffhunkinfoColor': QColor(166, 72, 72, 255),
    # 'diffhunkinfoPaper': QColor(255, 255, 255, 255),
    # 'diffunmodifiedColor': QColor(102, 102, 102, 255),
    # 'diffunmodifiedPaper': QColor(238, 238, 238, 255),

    'ioconsolePaper': QColor(255, 255, 255, 255),
    'ioconsoleColor': QColor(0, 0, 0, 255),
    'ioconsoleMarginStdoutColor': QColor(0, 0, 0, 255),
    'ioconsoleMarginStdinColor': QColor(51, 102, 255, 255),
    'ioconsoleMarginStderrColor': QColor(204, 51, 0, 255),
    'ioconsoleMarginIDEMsgColor': QColor(128, 128, 128, 255),

    'invalidInputPaper': QColor(255, 193, 204, 100),

    'headerLabelBGColor': QColor(255, 255, 255, 255),
    'headerLabelBorderColor': QColor(179, 175, 171, 255)}


_DEFAULT_APP_CSS = """
QStatusBar::item
{ border: 0px solid black }
QToolTip
{ color: black;
  font-size: 11px;
  border: 1px solid gray;
  background-color: #fff;
  padding: 2px; }
QTreeView
{ alternate-background-color: #eef0f1;
  background-color: #fff; }
QLineEdit
{ background-color: #fff; }
QComboBox
{ background-color: #fff;
  color: black; }
QComboBox QAbstractItemView
{ outline: 0px; }
QTextEdit
{ background-color: #fff; }
QPlainTextEdit
{ background-color: #fff; }
QListView
{ background-color: #fff; }

QStatusBar StatusBarFramedLabel
{ border-radius: 3px;
  padding: 4px;
  background-color: #fff;
  border: 1px solid #b3afab; }
QStatusBar StatusBarPixmapLabel
{ background: transparent; }
QStatusBar StatusBarPathLabel
{ border-radius: 3px;
  padding: 4px;
  background-color: #fff;
  border: 1px solid #b3afab; }
FramedLabel
{ border-radius: 3px;
  padding: 4px;
  background: transparent;
  border: 1px solid #b3afab; }
HeaderLabel
{ border-radius: 3px;
  padding: 4px;
  background-color: #fff;
  border: 1px solid #b3afab; }
FramedFitLabel
{ border-radius: 3px;
  padding: 4px;
  background: transparent;
  border: 1px solid #b3afab; }
HeaderFitLabel
{ border-radius: 3px;
  padding: 4px;
  background-color: #fff;
  border: 1px solid #b3afab; }
FramedFitPathLabel
{ border-radius: 3px;
  padding: 4px;
  background: transparent;
  border: 1px solid #b3afab; }
HeaderFitPathLabel
{ border-radius: 3px;
  padding: 4px;
  background-color: #fff;
  border: 1px solid #b3afab; }"""


_DEFAULT_CFLOW_SETTINGS = {
    'debug': False,
    'cfMonoFont': buildFont('Courier,12,-1,5,50,0,0,0,0,0') if isMac else
                  buildFont('Monospace,12,-1,5,50,0,0,0,0,0'),
    'badgeFont': buildFont('Courier,9,-1,5,50,0,0,0,0,0') if isMac else
                 buildFont('Monospace,9,-1,5,50,0,0,0,0,0'),

    'hCellPadding': 6,      # in pixels (left and right)
    'vCellPadding': 8,      # in pixels (top and bottom)
    'hTextPadding': 4,      # in pixels (left and right)
    'vTextPadding': 4,      # in pixels (top and bottom)

    # Scope header (file, decor, loops etc) paddings
    'hHeaderPadding': 5,
    'vHeaderPadding': 5,

    'vSpacer': 8,

    # Rounded rectangles radius for the return-like statements
    'minWidth': 100,

    'cfLineWidth': 2,        # Control flow line
    'boxLineWidth': 1,        # All boxes and their leads to CF
    'cfLineColor': QColor(132, 132, 132, 255),

    # Selection
    'selectColor': QColor(63, 81, 181, 255),
    'selectPenWidth': 3,

    # Code blocks and other statements
    'codeBlockBGColor': QColor(248, 248, 248, 255),
    'codeBlockFGColor': QColor(0, 0, 0, 255),
    'codeBlockBorderColor': QColor(150, 150, 150, 255),

    'breakBGColor': QColor(250, 227, 217, 255),
    'breakFGColor': QColor(0, 0, 0, 255),
    'breakBorderColor': QColor(150, 150, 150, 255),
    'breakHPadding': 4,
    'breakVPadding': 0,
    'breakRectRadius': 2,

    'continueBGColor': QColor(138, 198, 209, 255),
    'continueFGColor': QColor(0, 0, 0, 255),
    'continueBorderColor': QColor(150, 150, 150, 255),
    'continueHPadding': 4,
    'continueVPadding': 0,
    'continueRectRadius': 2,

    'returnBGColor': QColor(255, 247, 188, 255),
    'returnFGColor': QColor(0, 0, 0, 255),
    'returnBorderColor': QColor(150, 150, 150, 255),
    'returnRectRadius': 12,

    'raiseBGColor': QColor(219, 210, 210, 255),
    'raiseFGColor': QColor(0, 0, 0, 255),
    'raiseBorderColor': QColor(150, 150, 150, 255),

    'assertBGColor': QColor(210, 210, 210, 255),
    'assertFGColor': QColor(0, 0, 0, 255),
    'assertBorderColor': QColor(150, 150, 150, 255),

    'sysexitBGColor': QColor(255, 156, 156, 255),
    'sysexitFGColor': QColor(0, 0, 0, 255),
    'sysexitBorderColor': QColor(60, 60, 60, 255),

    'importBGColor': QColor(248, 248, 248, 255),
    'importFGColor': QColor(0, 0, 0, 255),
    'importBorderColor': QColor(150, 150, 150, 255),

    'ifBGColor': QColor(255, 247, 188, 255),
    'ifFGColor': QColor(0, 0, 0, 255),
    'ifBorderColor': QColor(150, 150, 150, 255),
    'ifYBranchTextColor': QColor(132, 132, 132, 255),
    'ifNBranchTextColor': QColor(132, 132, 132, 255),
    'ifWidth': 10,           # One if wing width

    # Badges
    'badgeBGColor': QColor(230, 230, 230, 255),
    'badgeFGColor': QColor(0, 0, 0, 255),
    'badgeLineWidth': 1,
    'badgeBorderColor': QColor(150, 150, 150, 255),
    'badgeHSpacing': 2,
    'badgeVSpacing': 1,
    'badgeRadius': 2,
    'badgeShift': 4,

    # Comments: leading, side & independent
    'commentBGColor': QColor(255, 255, 235, 255),
    'commentFGColor': QColor(90, 90, 90, 255),
    'commentBorderColor': QColor(150, 150, 150, 255),
    'commentCorner': 5,      # Top right comment corner
    'vHiddenCommentPadding': 3,
    'hHiddenCommentPadding': 3,
    'hiddenCommentBGColor': QColor(255, 255, 235, 255),
    'hiddenCommentBorderColor': QColor(150, 150, 150, 255),
    'hiddenCommentRectRadius': 2,
    'ifSideCommentVShift': 7,

    'mainLine': 25,

    # docstring border cannot be changed; it is a property of the scope
    'docstringBGColor': QColor(255, 255, 235, 255),
    'docstringFGColor': QColor(90, 90, 90, 255),

    'scopeRectRadius': 4,

    'fileScopeBGColor': QColor(255, 255, 255, 255),
    'fileScopeFGColor': QColor(0, 0, 0, 255),
    'fileScopeBorderColor': QColor(150, 150, 150, 255),

    'funcScopeBGColor': QColor(240, 240, 255, 255),
    'funcScopeFGColor': QColor(0, 0, 0, 255),
    'funcScopeBorderColor': QColor(150, 150, 150, 255),

    'classScopeBGColor': QColor(240, 255, 240, 255),
    'classScopeFGColor': QColor(0, 0, 0, 255),
    'classScopeBorderColor': QColor(150, 150, 150, 255),

    'forScopeBGColor': QColor(198, 219, 239, 255),
    'forScopeFGColor': QColor(0, 0, 0, 255),
    'forScopeBorderColor': QColor(150, 150, 150, 255),

    'whileScopeBGColor': QColor(198, 219, 239, 255),
    'whileScopeFGColor': QColor(0, 0, 0, 255),
    'whileScopeBorderColor': QColor(150, 150, 150, 255),

    'tryScopeBGColor': QColor(248, 248, 248, 255),
    'tryScopeFGColor': QColor(0, 0, 0, 255),
    'tryScopeBorderColor': QColor(150, 150, 150, 255),

    'decorScopeBGColor': QColor(240, 240, 240, 255),
    'decorScopeFGColor': QColor(0, 0, 0, 255),
    'decorScopeBorderColor': QColor(150, 150, 150, 255),

    'withScopeBGColor': QColor(248, 248, 248, 255),
    'withScopeFGColor': QColor(0, 0, 0, 255),
    'withScopeBorderColor': QColor(150, 150, 150, 255),

    'exceptScopeBGColor': QColor(210, 210, 210, 255),
    'exceptScopeFGColor': QColor(0, 0, 0, 255),
    'exceptScopeBorderColor': QColor(150, 150, 150, 255),
    'hiddenExceptBGColor': QColor(240, 240, 240, 255),
    'hiddenExceptBorderColor': QColor(150, 150, 150, 255),
    'hHiddenExceptPadding': 3,
    'vHiddenExceptPadding': 3,
    'hiddenExceptRectRadius': 3,

    'finallyScopeBGColor': QColor(248, 248, 248, 255),
    'finallyScopeFGColor': QColor(0, 0, 0, 255),
    'finallyScopeBorderColor': QColor(150, 150, 150, 255),

    'forElseScopeBGColor': QColor(198, 219, 239, 255),
    'forElseScopeFGColor': QColor(0, 0, 0, 255),
    'forElseScopeBorderColor': QColor(150, 150, 150, 255),

    'whileElseScopeBGColor': QColor(198, 219, 239, 255),
    'whileElseScopeFGColor': QColor(0, 0, 0, 255),
    'whileElseScopeBorderColor': QColor(150, 150, 150, 255),

    'tryElseScopeBGColor': QColor(248, 248, 248, 255),
    'tryElseScopeFGColor': QColor(0, 0, 0, 255),
    'tryElseScopeBorderColor': QColor(150, 150, 150, 255),

    # Groups
    'openGroupVSpacer': 3,
    'openGroupHSpacer': 3,
    'openGroupBGColor': QColor(245, 255, 255, 255),
    'openGroupFGColor': QColor(0, 0, 0, 255),
    'openGroupBorderColor': QColor(32, 32, 32, 255),
    'openGroupControlBGColor': QColor(197, 217, 249, 255),
    'openGroupControlBorderColor': QColor(140, 179, 242, 255),
    'openGroupControlLineWidth': 1,

    'collapsedGroupBGColor': QColor(245, 255, 255, 255),
    'collapsedGroupFGColor': QColor(0, 0, 0, 255),
    'collapsedGroupBorderColor': QColor(150, 150, 150, 255),
    'collapsedGroupXShift': 4,
    'collapsedGroupYShift': 4,

    'emptyGroupBGColor': QColor(245, 255, 255, 255),
    'emptyGroupFGColor': QColor(0, 0, 0, 255),
    'emptyGroupBorderColor': QColor(150, 150, 150, 255),
    'emptyGroupXShift': 4,
    'emptyGroupYShift': 4,

    # Rubber band selection
    'rubberBandBorderColor': QColor(63, 81, 181, 255),
    'rubberBandFGColor': QColor(182, 182, 182, 64),

    # Doc links
    'hDocLinkPadding': 3,   # in pixels (left and right)
    'vDocLinkPadding': 3,   # in pixels (top and bottom)
    'docLinkBGColor': QColor(219, 230, 246, 255),
    'docLinkFGColor': QColor(0, 0, 0, 255),
    'docLinkBorderColor': QColor(150, 150, 150, 255)}


class Skin:

    """Holds the definitions for a skin"""

    def __init__(self):
        # That's a trick to be able to implement getattr/setattr
        self.__dirName = None
        self.__userDirName = None
        self.__appCSS = None
        self.__values = {}
        self.__overrides = {}
        self.minTextZoom = None
        self.minCFlowZoom = None
        self.__reset()

    def __reset(self):
        """Resets all the values to the default"""
        # Note: python 3.5 and 3.6 deepcopy() behaves different. The 3.6
        # fails to copy QFont at all while 3.5 copies it improperly.
        # So to be 100% sure it works, here is a manual copying...
        self.__dirName = None
        self.__overrides = {}
        self.__userDirName = USER_SKIN_DIR + \
                             _DEFAULT_SKIN_SETTINGS['name'] + os.path.sep
        if not os.path.exists(self.__userDirName):
            self.__userDirName = None
        elif not os.path.isdir(self.__userDirName):
            self.__userDirName = None

        self.__values = {}
        for key, value in _DEFAULT_SKIN_SETTINGS.items():
            if isinstance(value, QFont):
                self.__values[key] = QFont(_DEFAULT_SKIN_SETTINGS[key])
            else:
                self.__values[key] = value

        for key, value in _DEFAULT_CFLOW_SETTINGS.items():
            if isinstance(value, QFont):
                self.__values[key] = QFont(_DEFAULT_CFLOW_SETTINGS[key])
            else:
                self.__values[key] = value

        self.__appCSS = deepcopy(_DEFAULT_APP_CSS)
        self.__applyOverrides()

        self.minTextZoom = self.__calculateMinTextZoom()
        self.minCFlowZoom = self.__calculateMinCFlowZoom()

    def getDir(self):
        """Provides the directory where the skin is coming from"""
        # Can be None for the default skin and for the custom user skin
        return self.__dirName

    def getUserDir(self):
        """Provides the location of the user skin directory"""
        # Can be the same as self.__dirName for the custom user skin
        return self.__userDirName

    def __getitem__(self, key):
        if key == 'appCSS':
            return self.__appCSS
        return self.__values[key]

    def __setitem__(self, key, value):
        logging.error('The generic skin parameters are immutable')

    @property
    def cflowSettings(self):
        """Provides the cflow settings dictionary"""
        # It is more than needed. The minimum is what is in the CFOW dict
        return self.__values

    @staticmethod
    def flush(fName, values):
        """Saves the values to the disk"""
        try:
            with open(fName, 'w', encoding=DEFAULT_ENCODING) as diskfile:
                json.dump(values, diskfile, indent=4, default=colorFontToJSON)
        except Exception as exc:
            logging.error('Error updating skin settings (to %s): %s',
                          fName, str(exc))

    @staticmethod
    def flushCFlow(fName, values):
        """Saves the cflow settings to the disk"""
        try:
            with open(fName, 'w', encoding=DEFAULT_ENCODING) as diskfile:
                json.dump(values, diskfile, indent=4, default=colorFontToJSON)
        except Exception as exc:
            logging.error('Error updating control flow settings (to %s): %s',
                          fName, str(exc))

    def loadByName(self, skinName):
        """Loads the skin by name.

        Implemetation should be in sync with getSkinsList
        """
        if self.__values['name'] == skinName:
            return

        if skinName == _DEFAULT_SKIN_SETTINGS['name']:
            self.__reset()
            return

        # check if it is from an installation package
        skinsAndDirs = getSkinsWithDirs()
        if skinName not in skinsAndDirs:
            logging.error('Skin "%s" is not found', skinName)
            return

        skinDir = skinsAndDirs[skinName]

        # load 3 files
        newAppCSS = self.__loadAppCSS(skinDir + 'app.css')
        newValues = self.__loadSkin(skinDir + 'skin.json')
        newCflowValues = self.__loadCFlow(skinDir + 'cflow.json')
        if newAppCSS is None or newValues is None or newCflowValues is None:
            logging.error('The current skin ("%s") has not been replaced '
                          'with skin "%s"', self.__values['name'], skinName)
            return

        # Apply new values
        self.__appCSS = newAppCSS
        self.__values = newValues
        self.__values.update(newCflowValues)

        # Set the skin directories
        if skinDir.startswith(USER_SKIN_DIR):
            # Pure user dir, unknown for the package
            self.__dirName = None
            self.__userDirName = skinDir
        else:
            self.__dirName = skinDir
            self.__userDirName = USER_SKIN_DIR + \
                                 self.__values['name'] + os.path.sep
            if not os.path.exists(self.__userDirName):
                self.__userDirName = None
            elif not os.path.isdir(self.__userDirName):
                self.__userDirName = None

        self.__applyOverrides()
        self.__checkMissedValues()

        self.minTextZoom = self.__calculateMinTextZoom()
        self.minCFlowZoom = self.__calculateMinCFlowZoom()

    def __applyOverrides(self):
        """Applies the skin overrides"""
        if self.__userDirName is None:
            return
        fName = self.__userDirName + OVERRIDE_FILE
        if not os.path.exists(fName):
            return
        if not os.path.isfile(fName):
            return

        try:
            with open(fName, 'r', encoding=DEFAULT_ENCODING) as diskfile:
                diskContent = json.load(diskfile,
                                        object_hook=colorFontFromJSON)
                diskContent, _ = self.__postProcessValues(diskContent)
            self.__overrides = diskContent
            self.__values.update(diskContent)
        except Exception as exc:
            logging.error('Error applying overrides on the skin "%s" '
                          'from %s: %s',self.__values['name'], fName, str(exc))

    def __checkMissedValues(self):
        """Checks if some values are missed in the skin"""
        expectedValues = set(_DEFAULT_SKIN_SETTINGS.keys()) | \
                         set(_DEFAULT_CFLOW_SETTINGS.keys())
        presentValues = set(self.__values.keys())
        diff = expectedValues - presentValues
        for val in diff:
            logging.error('The skin "%s" misses value %s. The values will be '
                          'taken from the default settings.',
                          self.__values['name'], val)
            if val in _DEFAULT_SKIN_SETTINGS:
                defaultVal = {val: _DEFAULT_SKIN_SETTINGS[val]}
            else:
                defaultVal = {val: _DEFAULT_CFLOW_SETTINGS[val]}
            defaultVal, _ = self.__postProcessValues(defaultVal)
            self.__values.update(defaultVal)

    @staticmethod
    def __loadAppCSS(fName):
        """Loads the application CSS file"""
        try:
            return getFileContent(fName)
        except Exception as exc:
            logging.error('Cannot read an application CSS from %s: %s',
                          fName, str(exc))
        return None

    def __loadSkin(self, fName):
        """Loads the general settings file"""
        try:
            with open(fName, 'r', encoding=DEFAULT_ENCODING) as diskfile:
                diskContent = json.load(diskfile,
                                        object_hook=colorFontFromJSON)
                diskContent, oldFormat = self.__postProcessValues(diskContent)
            if oldFormat:
                try:
                    self.flush(fName, diskContent)
                except:
                    pass
            return diskContent
        except Exception as exc:
            logging.error('Cannot read skin settings from %s: %s',
                          fName, str(exc))
        return None

    def __loadCFlow(self, fName):
        """Loads control flow settings file"""
        try:
            with open(fName, 'r', encoding=DEFAULT_ENCODING) as diskfile:
                diskContent = json.load(diskfile,
                                        object_hook=colorFontFromJSON)
                diskContent, oldFormat = self.__postProcessValues(diskContent)
            if oldFormat:
                try:
                    self.flushCFlow(fName, diskContent)
                except:
                    pass
            return diskContent
        except Exception as exc:
            logging.error('Cannot read control flow settings from %s: %s',
                          fName, str(exc))
        return None

    def __calculateMinTextZoom(self):
        """Calculates the minimum text zoom"""
        marginPointSize = self.__values['lineNumFont'].pointSize()
        mainAreaPointSize = self.__values['monoFont'].pointSize()
        return (min(marginPointSize, mainAreaPointSize) - 1) * -1

    def __calculateMinCFlowZoom(self):
        """Calculates the minimum control flow zoom"""
        badgePointSize = self.__values['badgeFont'].pointSize()
        monoPointSize = self.__values['cfMonoFont'].pointSize()
        return (min(badgePointSize, monoPointSize) - 1) * -1

    @staticmethod
    def __postProcessValues(values):
        """Builds fonts and colors as needed"""
        # Detection is primitive: the name of the dictionary item
        oldFormat = False
        for name, value in values.items():
            lowerName = name.lower()
            if 'font' in lowerName:
                if isinstance(value, QFont):
                    # already built
                    oldFormat = True
                else:
                    values[name] = buildFont(value)
            elif 'color' in lowerName or 'paper' in lowerName:
                if isinstance(value, QColor):
                    # already built
                    oldFormat = True
                else:
                    values[name] = buildColor(value)
        return values, oldFormat

    def __createUserSkinDir(self):
        """Creates the user skin dir if necessary"""
        if not self.__userDirName:
            self.__userDirName = USER_SKIN_DIR + self.__values['name'] + \
                                 os.path.sep
            if os.path.exists(self.__userDirName):
                if not os.path.isdir(self.__userDirName):
                    logging.error('The skin needs to have the directory %s '
                                  'to be available but the name is taken by '
                                  'something else. No skin overrides '
                                  'possible.', self.__userDirName)
                    self.__userDirName = None
                    return False
            # Try to create the dir
            try:
                os.makedirs(self.__userDirName, exist_ok=True)
            except Exception as exc:
                logging.error('Error creating the skin overriding '
                              'directory %s: %s', self.__userDirName, str(exc))
                self.__userDirName = None
                return False
        return True

    def __flushOverrides(self):
        """Saves the overrides to the disk"""
        # The user directory must exist
        fName = self.__userDirName + OVERRIDE_FILE
        try:
            with open(fName, 'w', encoding=DEFAULT_ENCODING) as diskfile:
                json.dump(self.__overrides, diskfile, indent=4,
                          default=colorFontToJSON)
        except Exception as exc:
            logging.error('Error updating overridden skin settings in %s: %s',
                          fName, str(exc))

    def setTextMonoFont(self, font):
        """Sets the new mono font family"""
        self.__values['monoFont'] = font
        self.__overrides['monoFont'] = font
        if self.__createUserSkinDir():
            self.__flushOverrides()

    def setMarginFont(self, font):
        """Sets the new mono font family"""
        self.__values['lineNumFont'] = font
        self.__overrides['lineNumFont'] = font
        if self.__createUserSkinDir():
            self.__flushOverrides()

    def setFlowMonoFont(self, font):
        """Sets the new flow font family"""
        self.__values['cfMonoFont'] = font
        self.__overrides['cfMonoFont'] = font
        if self.__createUserSkinDir():
            self.__flushOverrides()

    def setFlowBadgeFont(self, font):
        """Sets the new flow badge font"""
        self.__values['badgeFont'] = font
        self.__overrides['badgeFont'] = font
        if self.__createUserSkinDir():
            self.__flushOverrides()


def isSkinDir(dName):
    """True if all the required files are there"""
    for fName in ['app.css', 'skin.json', 'cflow.json']:
        if not os.path.exists(dName + fName):
            return False
    return True


def getSkinName(dName):
    """Provides the skin name or None"""
    try:
        fName = dName + 'skin.json'
        with open(fName, 'r',
                  encoding=DEFAULT_ENCODING) as diskfile:
            # Note: the load() method lacks the
            # object_hook=colorFontFromJSON parameter because the only
            # the skin name is read and no font/color conversion needed
            values = json.load(diskfile)
            return values['name']
    except:
        return None


# Dictionary { <skin name>: <directory> }
SKIN_LIST = None


def getSkinsList():
    """Provides just the skin names"""
    return getSkinsWithDirs().keys()


def getSkinsWithDirs():
    """Builds a list of skins - system wide and the user local"""
    global SKIN_LIST

    if SKIN_LIST is None:
        # default is coming from memory and always there
        SKIN_LIST = {'default': None}

        # First, walk the installation skin dirs
        for item in os.listdir(PACKAGE_SKIN_DIR):
            dName = PACKAGE_SKIN_DIR + item + os.path.sep
            if os.path.isdir(dName):
                if isSkinDir(dName):
                    name = getSkinName(dName)
                    if name and name not in SKIN_LIST:
                        SKIN_LIST[name] = dName

        for item in os.listdir(USER_SKIN_DIR):
            if item == SAMPLE_SKIN:
                continue
            if item in SKIN_LIST:
                # this is overriding of the IDE supplied skins
                continue
            dName = USER_SKIN_DIR + item + os.path.sep
            if os.path.isdir(dName):
                if isSkinDir(dName):
                    name = getSkinName(dName)
                    if name and name not in SKIN_LIST:
                        SKIN_LIST[name] = dName
    return SKIN_LIST


def populateSampleSkin():
    """Populates the sample skin in the user directory"""
    dName = USER_SKIN_DIR + SAMPLE_SKIN + os.path.sep
    if os.path.exists(dName):
        if not os.path.isdir(dName):
            logging.error('Error creating a sample skin at %s. '
                          'The file system entry is already occupied '
                          'by something else', dName)
            return

    # Try to create the dir
    try:
        os.makedirs(dName, exist_ok=True)
    except Exception as exc:
        logging.error('Error creating a sample skin at %s: %s',
                      dName, str(exc))

    fName = dName + 'app.css'
    try:
        saveToFile(fName, _DEFAULT_APP_CSS)
    except Exception as exc:
        logging.error('Error creating a sample skin at %s. '
                      'Cannot write app.css: %s', dName, str(exc))

    fName = dName + 'skin.json'
    try:
        values = {}
        for key, value in _DEFAULT_SKIN_SETTINGS.items():
            if isinstance(value, QFont):
                values[key] = QFont(_DEFAULT_SKIN_SETTINGS[key])
            else:
                values[key] = value
        values['name'] = SAMPLE_SKIN
        with open(fName, 'w', encoding=DEFAULT_ENCODING) as diskfile:
            json.dump(values, diskfile, indent=4, default=colorFontToJSON)
    except Exception as exc:
        logging.error('Error creating a sample skin at %s. '
                      'Cannot write skin.json: %s', dName, str(exc))

    fName = dName + 'cflow.json'
    try:
        values = {}
        for key, value in _DEFAULT_CFLOW_SETTINGS.items():
            if isinstance(value, QFont):
                values[key] = QFont(_DEFAULT_CFLOW_SETTINGS[key])
            else:
                values[key] = value
        with open(fName, 'w', encoding=DEFAULT_ENCODING) as diskfile:
            json.dump(values, diskfile, indent=4, default=colorFontToJSON)
    except Exception as exc:
        logging.error('Error creating a sample skin at %s. '
                      'Cannot write cflow.json: %s', dName, str(exc))

